#!/usr/bin/env python
# crate_anon/common/sql.py

import argparse
import datetime
import logging
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from sqlalchemy import inspect
import sqlalchemy.engine
import sqlalchemy.orm.session
import sqlalchemy.schema

log = logging.getLogger(__name__)


# =============================================================================
# SQL elements: literals, identifiers
# =============================================================================

def sql_string_literal(text: str) -> str:
    return "'" + text.replace("'", "''") + "'"


def sql_date_literal(dt: datetime.datetime) -> str:
    return dt.strftime("'%Y-%m-%d'")


def sql_datetime_literal(dt: datetime.datetime,
                         subsecond: bool = False) -> str:
    fmt = "'%Y-%m-%dT%H:%M:%S{}'".format(".%f" if subsecond else "")
    return dt.strftime(fmt)


def combine_db_table(db: str, table: str) -> str:
    if db:
        return "{}.{}".format(db, table)
    else:
        return table


def split_db_table(dbtable: str) -> Tuple[Optional[str], str]:
    components = dbtable.split('.')
    if len(components) == 2:  # db.table
        return components[0], components[1]
    elif len(components) == 1:  # table
        return None, components[0]
    else:
        raise ValueError("Bad dbtable: {}".format(dbtable))


# =============================================================================
# SQLAlchemy reflection and DDL
# =============================================================================

_print_not_execute = False


def set_print_not_execute(print_not_execute: bool) -> None:
    global _print_not_execute
    _print_not_execute = print_not_execute


def format_sql_for_print(sql: str) -> str:
    # Remove blank lines and trailing spaces
    lines = list(filter(None, [x.replace("\t", "    ").rstrip()
                               for x in sql.splitlines()]))
    # Shift all lines left if they're left-padded
    firstleftpos = float('inf')
    for line in lines:
        leftpos = len(line) - len(line.lstrip())
        firstleftpos = min(firstleftpos, leftpos)
    if firstleftpos > 0:
        lines = [x[firstleftpos:] for x in lines]
    return "\n".join(lines)


def sql_fragment_cast_to_int(expr: str) -> str:
    # Conversion to INT:
    # http://stackoverflow.com/questions/2000045
    # http://stackoverflow.com/questions/14719760  # this one
    # http://stackoverflow.com/questions/14692131
    return "CASE WHEN {expr} NOT LIKE '%[^0-9]%' " \
           "THEN CAST({expr} AS INTEGER) ELSE NULL END".format(expr=expr)


def execute(engine: sqlalchemy.engine.Engine, sql: str) -> None:
    log.debug(sql)
    if _print_not_execute:
        print(format_sql_for_print(sql) + "\n;")
        # extra \n in case the SQL ends in a comment
    else:
        engine.execute(sql)


def add_columns(engine: sqlalchemy.engine.Engine,
                table: sqlalchemy.schema.Table,
                name_coltype_dict: Dict[str,
                                        sqlalchemy.schema.Column]) -> None:
    existing_column_names = get_column_names(engine, tablename=table.name,
                                             to_lower=True)
    column_defs = []
    for name, coltype in name_coltype_dict.items():
        if name.lower() not in existing_column_names:
            column_defs.append("{} {}".format(name, coltype))
        else:
            log.debug("Table '{}': column '{}' already exists; not "
                      "adding".format(table.name, name))
    # ANSI SQL: add one column at a time: ALTER TABLE ADD [COLUMN] coldef
    #   - i.e. "COLUMN" optional, one at a time, no parentheses
    #   - http://www.contrib.andrew.cmu.edu/~shadow/sql/sql1992.txt
    # MySQL: ALTER TABLE ADD [COLUMN] (a INT, b VARCHAR(32));
    #   - i.e. "COLUMN" optional, parentheses required for >1, multiple OK
    #   - http://dev.mysql.com/doc/refman/5.7/en/alter-table.html
    # MS SQL Server: ALTER TABLE ADD COLUMN a INT, B VARCHAR(32);
    #   - i.e. no "COLUMN", no parentheses, multiple OK
    #   - https://msdn.microsoft.com/en-us/library/ms190238.aspx
    #   - https://msdn.microsoft.com/en-us/library/ms190273.aspx
    #   - http://stackoverflow.com/questions/2523676
    # SQLAlchemy doesn't provide a shortcut for this.
    for column_def in column_defs:
        log.info("Table '{}': adding column {}".format(
            table.name, column_def))
        execute(engine, """
            ALTER TABLE {tablename} ADD {column_def}
        """.format(tablename=table.name, column_def=column_def))


def drop_columns(engine: sqlalchemy.engine.Engine,
                 table: sqlalchemy.schema.Table,
                 column_names: Iterable[str]) -> None:
    existing_column_names = get_column_names(engine, tablename=table.name,
                                             to_lower=True)
    for name in column_names:
        if name.lower() not in existing_column_names:
            log.debug("Table '{}': column '{}' does not exist; not "
                      "dropping".format(table.name, name))
        else:
            log.info("Table '{}': dropping column '{}'".format(table.name,
                                                               name))
            sql = "ALTER TABLE {t} DROP COLUMN {c}".format(t=table.name,
                                                           c=name)
            # SQL Server:
            #   http://www.techonthenet.com/sql_server/tables/alter_table.php
            # MySQL:
            #   http://dev.mysql.com/doc/refman/5.7/en/alter-table.html
            execute(engine, sql)


def add_indexes(engine: sqlalchemy.engine.Engine,
                table: sqlalchemy.schema.Table,
                indexdictlist: Iterable[Dict[str, Any]]) -> None:
    existing_index_names = get_index_names(engine, tablename=table.name,
                                           to_lower=True)
    for idxdefdict in indexdictlist:
        index_name = idxdefdict['index_name']
        column = idxdefdict['column']
        if not isinstance(column, str):
            column = ", ".join(column)  # must be a list
        unique = idxdefdict.get('unique', False)
        if index_name.lower() not in existing_index_names:
            log.info("Table '{}': adding index '{}' on columns '{}'".format(
                table.name, index_name, column))
            execute(engine, """
              CREATE{unique} INDEX {idxname} ON {tablename} ({column})
            """.format(
                unique=" UNIQUE" if unique else "",
                idxname=index_name,
                tablename=table.name,
                column=column,
            ))
        else:
            log.debug("Table '{}': index '{}' already exists; not "
                      "adding".format(table.name, index_name))


def drop_indexes(engine: sqlalchemy.engine.Engine,
                 table: sqlalchemy.schema.Table,
                 index_names: Iterable[str]) -> None:
    existing_index_names = get_index_names(engine, tablename=table.name,
                                           to_lower=True)
    for index_name in index_names:
        if index_name.lower() not in existing_index_names:
            log.debug("Table '{}': index '{}' does not exist; not "
                      "dropping".format(table.name, index_name))
        else:
            log.info("Table '{}': dropping index '{}'".format(table.name,
                                                              index_name))
            if engine.dialect.name == 'mysql':
                sql = "ALTER TABLE {t} DROP INDEX {i}".format(t=table.name,
                                                              i=index_name)
            elif engine.dialect.name == 'mssql':
                sql = "DROP INDEX {t}.{i}".format(t=table.name, i=index_name)
            else:
                assert False, "Unknown dialect: {}".format(engine.dialect.name)
            execute(engine, sql)


def get_table_names(engine: sqlalchemy.engine.Engine,
                    to_lower: bool = False,
                    sort: bool = False) -> List[str]:
    inspector = inspect(engine)
    table_names = inspector.get_table_names()
    if to_lower:
        table_names = [x.lower() for x in table_names]
    if sort:
        table_names = sorted(table_names, key=lambda x: x.lower())
    return table_names


def get_view_names(engine: sqlalchemy.engine.Engine,
                   to_lower: bool = False,
                   sort: bool = False) -> List[str]:
    inspector = inspect(engine)
    view_names = inspector.get_view_names()
    if to_lower:
        view_names = [x.lower() for x in view_names]
    if sort:
        view_names = sorted(view_names, key=lambda x: x.lower())
    return view_names


def get_column_names(engine: sqlalchemy.engine.Engine,
                     tablename: str,
                     to_lower: bool = False,
                     sort: bool = False) -> List[str]:
    """
    Reads columns names afresh from the database (in case metadata is out of
    date).
    """
    inspector = inspect(engine)
    columns = inspector.get_columns(tablename)
    column_names = [x['name'] for x in columns]
    if to_lower:
        column_names = [x.lower() for x in column_names]
    if sort:
        column_names = sorted(column_names, key=lambda x: x.lower())
    return column_names


def get_index_names(engine: sqlalchemy.engine.Engine,
                    tablename: str,
                    to_lower: bool = False) -> List[str]:
    """
    Reads index names from the database.
    """
    # http://docs.sqlalchemy.org/en/latest/core/reflection.html
    inspector = inspect(engine)
    indexes = inspector.get_indexes(tablename)
    index_names = [x['name'] for x in indexes if x['name']]
    # ... at least for SQL Server, there always seems to be a blank one
    # with {'name': None, ...}.
    if to_lower:
        index_names = [x.lower() for x in index_names]
    return index_names


def ensure_columns_present(engine: sqlalchemy.engine.Engine,
                           tablename: str,
                           column_names: Iterable[str]) -> None:
    existing_column_names = get_column_names(engine, tablename=tablename,
                                             to_lower=True)
    if not column_names:
        return
    for col in column_names:
        if col.lower() not in existing_column_names:
            raise ValueError(
                "Column '{}' missing from table '{}'".format(col, tablename))


def create_view(engine: sqlalchemy.engine.Engine,
                viewname: str,
                select_sql: str) -> None:
    # MySQL has CREATE OR REPLACE VIEW.
    # SQL Server doesn't: http://stackoverflow.com/questions/18534919
    if engine.dialect.name == 'mysql':
        sql = "CREATE OR REPLACE VIEW {viewname} AS {select_sql}".format(
            viewname=viewname,
            select_sql=select_sql,
        )
    else:
        drop_view(engine, viewname, quiet=True)
        sql = "CREATE VIEW {viewname} AS {select_sql}".format(
            viewname=viewname,
            select_sql=select_sql,
        )
    log.info("Creating view: '{}'".format(viewname))
    execute(engine, sql)


def drop_view(engine: sqlalchemy.engine.Engine,
              viewname: str,
              quiet: bool = False) -> None:
    # MySQL has DROP VIEW IF EXISTS, but SQL Server only has that from
    # SQL Server 2016 onwards.
    # - https://msdn.microsoft.com/en-us/library/ms173492.aspx
    # - http://dev.mysql.com/doc/refman/5.7/en/drop-view.html
    view_names = get_view_names(engine, to_lower=True)
    if viewname.lower() not in view_names:
        log.debug("View {} does not exist; not dropping".format(viewname))
    else:
        if not quiet:
            log.info("Dropping view: '{}'".format(viewname))
        sql = "DROP VIEW {viewname}".format(viewname=viewname)
        execute(engine, sql)


# =============================================================================
# View-building assistance class
# =============================================================================

class ViewMaker(object):
    def __init__(self,
                 engine: sqlalchemy.engine.Engine,
                 basetable: str,
                 existing_to_lower: bool = False,
                 rename: bool = None,
                 progargs: argparse.Namespace = None) -> None:
        rename = rename or {}
        self.engine = engine
        self.basetable = basetable
        self.progargs = progargs  # only for others' benefit
        self.select_elements = []
        for colname in get_column_names(engine, tablename=basetable,
                                        to_lower=existing_to_lower):
            if colname in rename:
                rename_to = rename[colname]
                if not rename_to:
                    continue
                as_clause = " AS {}".format(rename_to)
            else:
                as_clause = ""
            self.select_elements.append("{t}.{c}{as_clause}".format(
                t=basetable, c=colname, as_clause=as_clause))
        assert self.select_elements, "Must have some active SELECT elements " \
                                     "from base table"
        self.from_elements = [basetable]
        self.where_elements = []
        self.lookup_table_keyfields = []  # of (table, keyfield(s)) tuples

    def add_select(self, clause: str) -> None:
        self.select_elements.append(clause)

    def add_from(self, clause: str) -> None:
        self.from_elements.append(clause)

    def add_where(self, clause: str) -> None:
        self.where_elements.append(clause)

    def get_sql(self) -> str:
        if self.where_elements:
            where = "\n    WHERE {}".format(
                "\n        AND ".join(self.where_elements))
        else:
            where = ""
        return (
            "\n    SELECT {select_elements}"
            "\n    FROM {from_elements}{where}".format(
                select_elements=",\n        ".join(self.select_elements),
                from_elements="\n        ".join(self.from_elements),
                where=where))

    def record_lookup_table_keyfield(
            self,
            table: str,
            keyfield: Union[str, Iterable[str]]) -> None:
        self.lookup_table_keyfields.append((table, keyfield))

    def record_lookup_table_keyfields(
            self,
            table_keyfield_tuples: Iterable[
                Tuple[str, Union[str, Iterable[str]]]
            ]) -> None:
        for t, k in table_keyfield_tuples:
            self.record_lookup_table_keyfield(t, k)

    def get_lookup_tables(self) -> List[str]:
        return list(set(table for table, keyfield
                        in self.lookup_table_keyfields))

    def get_lookup_table_keyfields(self) -> List[Tuple[str, str]]:
        return list(self.lookup_table_keyfields)


# =============================================================================
# Transaction size-limiting class
# =============================================================================

class TransactionSizeLimiter(object):
    def __init__(self, session: sqlalchemy.orm.session.Session,
                 max_rows_before_commit: int = None,
                 max_bytes_before_commit: int = None) -> None:
        self._session = session
        self._max_rows_before_commit = max_rows_before_commit
        self._max_bytes_before_commit = max_bytes_before_commit
        self._bytes_in_transaction = 0
        self._rows_in_transaction = 0

    def commit(self) -> None:
        self._session.commit()
        self._bytes_in_transaction = 0
        self._rows_in_transaction = 0

    def notify(self, n_rows: int, n_bytes: int,
               force_commit: bool=False) -> None:
        if force_commit:
            self.commit()
            return
        self._bytes_in_transaction += n_bytes
        self._rows_in_transaction += n_rows
        # log.critical(
        #     "adding {} rows, {} bytes, "
        #     "to make {} rows, {} bytes so far".format(
        #         n_rows, n_bytes,
        #         self._rows_in_transaction, self._bytes_in_transaction))
        if ((self._max_bytes_before_commit is not None and
                self._bytes_in_transaction >= self._max_bytes_before_commit) or
                (self._max_rows_before_commit is not None and
                 self._rows_in_transaction >= self._max_rows_before_commit)):
            log.info("Triggering early commit based on row/byte count")
            self.commit()



#!/usr/bin/env python
# crate_anon/crateweb/consent/utils.py

from functools import lru_cache
import os
from typing import Any, Dict

from django.conf import settings
from django.core.exceptions import ValidationError
from django.template.loader import render_to_string


def read_static_file_contents(filename: str) -> str:
    with open(os.path.join(settings.LOCAL_STATIC_DIR, filename)) as f:
        return f.read()


def pdf_css(patient: bool = True) -> str:
    contents = read_static_file_contents('base.css')
    context = {
        'fontsize': (settings.PATIENT_FONTSIZE
                     if patient else settings.RESEARCHER_FONTSIZE),
    }
    contents += render_to_string('pdf.css', context)
    return contents


@lru_cache(maxsize=None)
def pdf_template_dict(patient: bool = True) -> Dict[str, str]:
    return {
        'css': pdf_css(patient),
        'PDF_LOGO_ABS_URL': settings.PDF_LOGO_ABS_URL,
        'PDF_LOGO_WIDTH': settings.PDF_LOGO_WIDTH,
    }


def render_pdf_html_to_string(template: str,
                              context: Dict[str, Any] = None,
                              patient: bool = True) -> str:
    context = context or {}
    context.update(pdf_template_dict(patient))
    return render_to_string(template, context)


def email_css() -> str:
    contents = read_static_file_contents('base.css')
    contents += render_to_string('email.css')
    return contents


@lru_cache(maxsize=None)
def email_template_dict() -> Dict[str, str]:
    return {
        'css': email_css(),
    }


def render_email_html_to_string(template: str,
                                context: Dict[str, Any] = None) -> str:
    context = context or {}
    context.update(email_template_dict())
    return render_to_string(template, context)


def get_domain_from_email(email: str) -> str:
    # Very simple version...
    try:
        return email.split('@')[1]
    except:
        raise ValidationError("Bad e-mail address: no domain")


def validate_researcher_email_domain(email: str) -> None:
    if not settings.VALID_RESEARCHER_EMAIL_DOMAINS:
        # Anything goes.
        return
    domain = get_domain_from_email(email)
    for valid_domain in settings.VALID_RESEARCHER_EMAIL_DOMAINS:
        if domain.lower() == valid_domain.lower():
            return
    raise ValidationError("Invalid researcher e-mail domain")



#!/usr/bin/env python
# crate_anon/nlp_manager/base_nlp_parser.py

# Simple base class for all our NLP parsers (GATE, regex, ...)

from functools import lru_cache
import logging
import sys
from typing import Any, Dict, Iterator, List, Optional, Tuple

from sqlalchemy.schema import Column, Index, Table
from sqlalchemy.sql import and_, exists, or_
from sqlalchemy.types import BigInteger

from crate_anon.common.timing import timer
from crate_anon.nlp_manager.constants import (
    FN_SRCPKVAL,
    FN_SRCPKSTR,
    SqlTypeDbIdentifier,
)
from crate_anon.nlp_manager.input_field_config import InputFieldConfig

# if sys.version_info.major >= 3 and sys.version_info.minor >= 5:
#     from crate_anon.nlp_manager import nlp_definition  # see PEP0484
from crate_anon.nlp_manager.nlp_definition import NlpDefinition

log = logging.getLogger(__name__)

TIMING_INSERT = "BaseNlpParser_process_sql_insert"
TIMING_PARSE = "parse"


# =============================================================================
# Base class for all parser types
# =============================================================================

class BaseNlpParser(object):
    FN_PK = '_pk'
    FN_NLPDEF = '_nlpdef'

    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        self._nlpdef = nlpdef
        self._cfgsection = cfgsection
        self._commit = commit
        self._destdb_name = None
        self._destdb = None
        if nlpdef is not None:
            self._destdb_name = nlpdef.opt_str(cfgsection, 'destdb',
                                               required=True)
            self._destdb = nlpdef.get_database(self._destdb_name)

    @classmethod
    def _core_dest_columns(cls) -> List[Column]:
        return [
            Column(cls.FN_PK, BigInteger, primary_key=True,
                   autoincrement=True,
                   doc="Arbitrary PK of output record"),
            Column(cls.FN_NLPDEF, SqlTypeDbIdentifier,
                   doc="Name of the NLP definition producing this row"),
        ]

    def dest_tables_columns(self) -> Dict[str, List[Column]]:
        """
        Returns a dictionary of {tablename: destination_columns}.
        """
        raise NotImplementedError

    def dest_tables_indexes(self) -> Dict[str, List[Index]]:
        return {}

    def get_metadata(self):
        return self._destdb.metadata

    def get_session(self):
        return self._destdb.session

    def get_engine(self):
        return self._destdb.engine

    def get_nlpdef_name(self) -> str:
        if self._nlpdef is None:
            return None
        return self._nlpdef.get_name()

    def get_parser_name(self) -> str:
        return getattr(self, 'NAME', None)

    def get_dbname(self) -> str:
        return self._destdb_name

    @staticmethod
    def _assert_no_overlap(description1: str, cols1: List[Column],
                           description2: str, cols2: List[Column]) -> None:
        """Used for ensuring non-overlapping column names."""
        set1 = set(c.name for c in cols1)
        set2 = set(c.name for c in cols2)
        assert not (set1 & set2), (
            "Overlap between {} column names ({}) and {} column names "
            "({})".format(description1, set1, description2, set2)
        )

    @staticmethod
    def _assert_column_lists_identical(
            list_of_column_lists: List[List[Any]],
            description: str) -> None:
        n = len(list_of_column_lists)
        if n <= 1:
            return
        for i in range(n - 1):
            if list_of_column_lists[i] != list_of_column_lists[i + 1]:
                msg = (
                    "Mismatch between {} lists: {}. (Are you trying to blend "
                    "source tables with different column names into a single "
                    "NLP results table?".format(
                        description, list_of_column_lists)
                )
                log.critical(msg)
                raise ValueError(msg)

    @lru_cache(maxsize=None)
    def tables(self) -> Dict[str, Table]:
        """
        Returns a dictionary of {tablename: Table}.
        """
        # Obtain a single set of copy columns
        ifconfigs = self._nlpdef.get_ifconfigs()
        assert ifconfigs, "Must specify a list of InputFieldConfigs"
        assert self._destdb, "Cannot use tables() call without a database"

        copycolumns_list = [i.get_copy_columns() for i in ifconfigs]
        self._assert_column_lists_identical(copycolumns_list, "column")
        copy_columns = copycolumns_list[0]

        source_columns = InputFieldConfig.get_srcref_columns_for_dest()
        self._assert_no_overlap("copy", copy_columns,
                                "source", source_columns)

        # Create one or more tables
        meta = self.get_metadata()
        tables = {}
        t_columns = self.dest_tables_columns()
        for tablename, destcols in t_columns.items():
            self._assert_no_overlap("copy", copy_columns,
                                    "destination", destcols)
            # And to check we haven't introduced any bugs internally:
            self._assert_no_overlap("source", source_columns,
                                    "destination", destcols)

            columns = (self._core_dest_columns() +
                       source_columns +
                       destcols +
                       copy_columns)
            copy_of_cols = [c.copy() for c in columns]

            t_indexes = self.dest_tables_indexes()
            dest_indexes = []
            if tablename in t_indexes:
                dest_indexes = t_indexes[tablename]
            copyindexes_list = [i.get_copy_indexes() for i in ifconfigs]
            self._assert_column_lists_identical(copyindexes_list, "index")
            copy_indexes = copyindexes_list[0]
            source_indexes = InputFieldConfig.get_srcref_indexes_for_dest()

            column_like_things = (
                copy_of_cols + source_indexes + dest_indexes + copy_indexes)
            # log.critical(repr(column_like_things))
            tables[tablename] = Table(tablename, meta, *column_like_things)
            # You can put indexes in the column list:
            # http://docs.sqlalchemy.org/en/latest/core/constraints.html

            # NOTE that after creating the Table, all the column objects get
            # "contaminated" by the link to it, so you have to start afresh
            # with new column objects, or take a further copy, as above.

            # You can copy a Column, but not an Index.
        return tables

    def get_tablenames(self) -> List[str]:
        return self.dest_tables_columns().keys()

    def get_table(self, tablename: str) -> Table:
        tables = self.tables()
        assert tablename in tables
        return tables[tablename]

    def make_tables(self, drop_first: bool = False) -> List[str]:
        assert self._destdb, "Cannot use tables() call without a database"
        engine = self.get_engine()
        tables = self.tables()
        pretty_names = []
        for t in tables.values():
            pretty_name = "{}.{}".format(self._destdb.name, t.name)
            if drop_first:
                log.info("Dropping table {}".format(pretty_name))
                t.drop(engine, checkfirst=True)
            log.info("Creating table {} (with indexes)".format(pretty_name))
            t.create(engine, checkfirst=True)
            pretty_names.append(pretty_name)
        return pretty_names

    def parse(self, text: str) -> Iterator[Tuple[str, Dict[str, Any]]]:
        """
        Takes the raw text as input.
        Yields (tablename, valuedict) tuples, where valuedict is
        a dictionary of {column: value}. The values returned are ONLY those
        generated by NLP, and do not include either (a) the source reference
        values (_srcdb, _srctable, etc.) or the "copy" fields.
        """
        raise NotImplementedError

    def process(self, text: str,
                starting_fields_values: Dict[str, Any]) -> None:
        starting_fields_values[self.FN_NLPDEF] = self._nlpdef.get_name()
        session = self.get_session()
        n_values = 0
        timer.start(TIMING_PARSE)
        for tablename, nlp_values in self.parse(text):
            timer.stop(TIMING_PARSE)
            # Merge dictionaries so EXISTING FIELDS/VALUES
            # (starting_fields_values) HAVE PRIORITY.
            nlp_values.update(starting_fields_values)
            sqla_table = self.get_table(tablename)
            # If we have superfluous keys in our dictionary, SQLAlchemy will
            # choke ("Unconsumed column names", reporting the thing that's
            # in our dictionary that it doesn't know about).
            # HOWEVER, note that SQLA column names may be mixed case (e.g.
            # 'Text') while our copy-column names are lower case (e.g. 'text'),
            # so we must have pre-converted the SQLA column names to lower
            # case. That happens in InputFieldConfig.get_copy_columns and
            # InputFieldConfig.get_copy_indexes
            column_names = [c.name for c in sqla_table.columns]
            final_values = {k: v for k, v in nlp_values.items()
                            if k in column_names}
            # log.critical(repr(sqla_table))
            insertquery = sqla_table.insert().values(final_values)
            timer.start(TIMING_INSERT)
            session.execute(insertquery)
            timer.stop(TIMING_INSERT)
            self._nlpdef.notify_transaction(
                session, n_rows=1, n_bytes=sys.getsizeof(final_values),
                force_commit=self._commit)  # or we get deadlocks in multiprocess mode  # noqa
            n_values += 1
            timer.start(TIMING_PARSE)
        timer.stop(TIMING_PARSE)
        log.debug("NLP processor {}/{}: found {} values".format(
            self.get_nlpdef_name(), self.get_parser_name(), n_values))

    def test(self):
        pass

    def test_parser(self, test_strings: List[str]) -> None:
        print("Testing parser: {}".format(type(self).__name__))
        for text in test_strings:
            print("    {} -> {}".format(text, list(self.parse(text))))

    def delete_dest_record(self,
                           ifconfig: InputFieldConfig,
                           srcpkval: int,
                           srcpkstr: Optional[str],
                           commit: bool = False) -> None:
        """
        Used during incremental updates.
        For when a record (specified by srcpkval) has been updated in the
        source; wipe older entries for it in the destination database(s).
        """
        session = self.get_session()
        srcdb = ifconfig.get_srcdb()
        srctable = ifconfig.get_srctable()
        srcpkfield = ifconfig.get_srcpkfield()
        destdb_name = self._destdb.name
        nlpdef_name = self._nlpdef.get_name()
        for tablename, desttable in self.tables().items():
            log.debug(
                "delete_from_dest_dbs... {}.{} -> {}.{}".format(
                    srcdb, srctable, destdb_name, tablename))
            # noinspection PyProtectedMember
            delquery = (
                desttable.delete().
                where(desttable.c._srcdb == srcdb).
                where(desttable.c._srctable == srctable).
                where(desttable.c._srcpkfield == srcpkfield).
                where(desttable.c._srcpkval == srcpkval).
                where(desttable.c._nlpdef == nlpdef_name)
            )
            if srcpkstr is not None:
                # noinspection PyProtectedMember
                delquery = delquery.where(desttable.c._srcpkstr == srcpkstr)
            session.execute(delquery)
            if commit:
                self._nlpdef.commit(session)
                # ... or we get deadlocks in incremental updates
                # http://dev.mysql.com/doc/refman/5.5/en/innodb-deadlocks.html

    def delete_where_srcpk_not(self,
                               ifconfig: InputFieldConfig,
                               temptable: Optional[Table]) -> None:
        destsession = self.get_session()
        srcdb = ifconfig.get_srcdb()
        srctable = ifconfig.get_srctable()
        for desttable_name, desttable in self.tables().items():
            log.debug(
                "delete_where_srcpk_not... {}.{} -> {}.{}".format(
                    srcdb, srctable, self._destdb_name, desttable_name))
            # noinspection PyProtectedMember
            dest_deletion_query = (
                desttable.delete().
                where(desttable.c._srcdb == srcdb).
                where(desttable.c._srctable == srctable).
                where(desttable.c._srcpkfield == ifconfig.get_srcpkfield()).
                where(desttable.c._nlpdef == self._nlpdef.get_name())
            )
            if temptable is not None:
                log.debug("... deleting selectively")
                #   DELETE FROM a WHERE NOT EXISTS (
                #       SELECT 1 FROM b
                #       WHERE a.a1 = b.b1
                #       AND (
                #           a.a2 = b.b2
                #           OR (a.a2 IS NULL AND b.b2 IS NULL)
                #       )
                #   )
                temptable_pkvalcol = temptable.columns[FN_SRCPKVAL]
                temptable_pkstrcol = temptable.columns[FN_SRCPKSTR]
                # noinspection PyProtectedMember
                dest_deletion_query = dest_deletion_query.where(
                    ~exists().where(
                        and_(
                            desttable.c._srcpkval == temptable_pkvalcol,
                            or_(
                                desttable.c._srcpkstr == temptable_pkstrcol,
                                and_(
                                    desttable.c._srcpkstr.is_(None),
                                    temptable_pkstrcol.is_(None)
                                )
                            )
                        )
                    )
                )
            else:
                log.debug("... deleting all")
            destsession.execute(dest_deletion_query)
            self._nlpdef.commit(destsession)



#!/usr/bin/env python
# crate_anon/crateweb/manage.py

import logging
import os
import shlex
import sys
from typing import List

import django
from django.core.management import execute_from_command_line

from crate_anon.crateweb.config.constants import CHERRYPY_EXTRA_ARGS_ENV_VAR

log = logging.getLogger(__name__)


os.environ.setdefault("DJANGO_SETTINGS_MODULE",
                      "crate_anon.crateweb.config.settings")

# from crate_anon.crateweb.config.settings import MIDDLEWARE_CLASSES
# print("1. MIDDLEWARE_CLASSES: {}".format(id(MIDDLEWARE_CLASSES)))
# print("1. MIDDLEWARE_CLASSES: {}".format(MIDDLEWARE_CLASSES))
django.setup()
# from crate_anon.crateweb.config.settings import MIDDLEWARE_CLASSES
# print("2. MIDDLEWARE_CLASSES: {}".format(id(MIDDLEWARE_CLASSES)))
# print("2. MIDDLEWARE_CLASSES: {}".format(MIDDLEWARE_CLASSES))

# print("sys.path: {}".format(sys.path))
# print("os.environ['DJANGO_SETTINGS_MODULE']: {}".format(
#     os.environ['DJANGO_SETTINGS_MODULE']))
# print("os.environ['{}']: {}".format(
#     CRATEWEB_CONFIG_ENV_VAR, os.environ[CRATEWEB_CONFIG_ENV_VAR]))


def main(argv: List[str] = None) -> None:
    if argv is None:
        argv = sys.argv
    # print(argv)
    execute_from_command_line(argv)


def runserver() -> None:
    argv = sys.argv[:]  # copy
    argv.insert(1, 'runserver')
    main(argv)


def runcpserver() -> None:
    argv = sys.argv[:]  # copy
    argv.insert(1, 'runcpserver')
    extraargs = shlex.split(os.environ.get(CHERRYPY_EXTRA_ARGS_ENV_VAR, ''))
    # log.critical(extraargs)
    argv.extend(extraargs)
    main(argv)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# crate_anon/nlp_manager/parse_clinical.py

import logging
from typing import Any, Dict, Iterator, List, Optional, Tuple

from sqlalchemy import Column, Integer, Float, String, Text

from crate_anon.nlp_manager.nlp_definition import NlpDefinition
from crate_anon.nlp_manager.regex_parser import (
    BaseNlpParser,
    common_tense,
    compile_regex,
    NumericalResultParser,
    OPTIONAL_RESULTS_IGNORABLES,
    PAST,
    PRESENT,
    RELATION,
    SimpleNumericalResultParser,
    TENSE_INDICATOR,
    to_float,
    to_pos_float,
    ValidatorBase,
    WORD_BOUNDARY,
)
from crate_anon.nlp_manager.regex_numbers import SIGNED_FLOAT
from crate_anon.nlp_manager.regex_units import (
    CM,
    FEET,
    INCHES,
    KG,
    kg_from_st_lb_oz,
    KG_PER_SQ_M,
    LB,
    M,
    m_from_ft_in,
    m_from_m_cm,
    MM_HG,
    STONES,
)

log = logging.getLogger(__name__)


# =============================================================================
#  Anthropometrics
# =============================================================================

# -----------------------------------------------------------------------------
# Height
# -----------------------------------------------------------------------------

class Height(NumericalResultParser):
    """Height. Handles metric and imperial."""
    METRIC_HEIGHT = r"""
        (                           # capture group
            (?:
                ( {SIGNED_FLOAT} )          # capture group
                {OPTIONAL_RESULTS_IGNORABLES}
                {M}
                {OPTIONAL_RESULTS_IGNORABLES}
                ( {SIGNED_FLOAT} )          # capture group
                {OPTIONAL_RESULTS_IGNORABLES}
                {CM}
            )
            | (?:
                ( {SIGNED_FLOAT} )          # capture group
                {OPTIONAL_RESULTS_IGNORABLES}
                {M}
            )
            | (?:
                ( {SIGNED_FLOAT} )          # capture group
                {OPTIONAL_RESULTS_IGNORABLES}
                {CM}
            )
        )
    """.format(SIGNED_FLOAT=SIGNED_FLOAT,
               OPTIONAL_RESULTS_IGNORABLES=OPTIONAL_RESULTS_IGNORABLES,
               M=M,
               CM=CM)
    IMPERIAL_HEIGHT = r"""
        (                           # capture group
            (?:
                ( {SIGNED_FLOAT} )      # capture group
                {OPTIONAL_RESULTS_IGNORABLES}
                {FEET}
                {OPTIONAL_RESULTS_IGNORABLES}
                ( {SIGNED_FLOAT} )      # capture group
                {OPTIONAL_RESULTS_IGNORABLES}
                {INCHES}
            )
            | (?:
                ( {SIGNED_FLOAT} )      # capture group
                {OPTIONAL_RESULTS_IGNORABLES}
                {FEET}
            )
            | (?:
                ( {SIGNED_FLOAT} )      # capture group
                {OPTIONAL_RESULTS_IGNORABLES}
                {INCHES}
            )
        )
    """.format(SIGNED_FLOAT=SIGNED_FLOAT,
               OPTIONAL_RESULTS_IGNORABLES=OPTIONAL_RESULTS_IGNORABLES,
               FEET=FEET,
               INCHES=INCHES)
    HEIGHT = r"""
        (?:
            \b height \b
        )
    """
    REGEX = r"""
        ( {HEIGHT} )                       # group for "BMI" or equivalent
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {TENSE_INDICATOR} )?             # optional group for tense indicator
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {RELATION} )?                    # optional group for relation
        {OPTIONAL_RESULTS_IGNORABLES}
        (?:
            {METRIC_HEIGHT}
            | {IMPERIAL_HEIGHT}
        )
    """.format(
        HEIGHT=HEIGHT,
        OPTIONAL_RESULTS_IGNORABLES=OPTIONAL_RESULTS_IGNORABLES,
        TENSE_INDICATOR=TENSE_INDICATOR,
        RELATION=RELATION,
        METRIC_HEIGHT=METRIC_HEIGHT,
        IMPERIAL_HEIGHT=IMPERIAL_HEIGHT,
        SIGNED_FLOAT=SIGNED_FLOAT,
        KG_PER_SQ_M=KG_PER_SQ_M,
    )

    COMPILED_REGEX = compile_regex(REGEX)
    NAME = "Height"
    PREFERRED_UNIT_COLUMN = "value_m"

    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False,
                 debug: bool = False) -> None:
        super().__init__(
            nlpdef=nlpdef,
            cfgsection=cfgsection,
            variable=self.NAME,
            target_unit=self.PREFERRED_UNIT_COLUMN,
            commit=commit
        )
        if debug:
            print("Regex for {}: {}".format(type(self).__name__, self.REGEX))

    def parse(self, text: str,
              debug: bool = False) -> Iterator[Tuple[str, Dict[str, Any]]]:
        for m in self.COMPILED_REGEX.finditer(text):  # watch out: 'm'/metres
            if debug:
                print("Match {} for {}".format(m, repr(text)))
            startpos = m.start()
            endpos = m.end()
            matching_text = m.group(0)  # the whole thing
            variable_text = m.group(1)
            tense_indicator = m.group(2)
            relation = m.group(3)
            metric_expression = m.group(4)
            metric_m_and_cm_m = m.group(5)
            metric_m_and_cm_cm = m.group(6)
            metric_m_only_m = m.group(7)
            metric_cm_only_cm = m.group(8)
            imperial_expression = m.group(9)
            imperial_ft_and_in_ft = m.group(10)
            imperial_ft_and_in_in = m.group(11)
            imperial_ft_only_ft = m.group(12)
            imperial_in_only_in = m.group(13)

            expression = None
            value_m = None
            if metric_expression:
                expression = metric_expression
                if metric_m_and_cm_m and metric_m_and_cm_cm:
                    metres = to_pos_float(metric_m_and_cm_m)
                    # ... beware: 'm' above
                    cm = to_pos_float(metric_m_and_cm_cm)
                    value_m = m_from_m_cm(metres=metres, centimetres=cm)
                elif metric_m_only_m:
                    value_m = to_pos_float(metric_m_only_m)
                elif metric_cm_only_cm:
                    cm = to_pos_float(metric_cm_only_cm)
                    value_m = m_from_m_cm(centimetres=cm)
            elif imperial_expression:
                expression = imperial_expression
                if imperial_ft_and_in_ft and imperial_ft_and_in_in:
                    ft = to_pos_float(imperial_ft_and_in_ft)
                    inches = to_pos_float(imperial_ft_and_in_in)
                    value_m = m_from_ft_in(feet=ft, inches=inches)
                elif imperial_ft_only_ft:
                    ft = to_pos_float(imperial_ft_only_ft)
                    value_m = m_from_ft_in(feet=ft)
                elif imperial_in_only_in:
                    inches = to_pos_float(imperial_in_only_in)
                    value_m = m_from_ft_in(inches=inches)

            tense, relation = common_tense(tense_indicator, relation)

            yield self.tablename, {
                self.FN_VARIABLE_NAME: self.variable,
                self.FN_CONTENT: matching_text,
                self.FN_START: startpos,
                self.FN_END: endpos,
                self.FN_VARIABLE_TEXT: variable_text,
                self.FN_RELATION: relation,
                self.FN_VALUE_TEXT: expression,
                self.target_unit: value_m,
                self.FN_TENSE: tense,
            }

    def test(self):
        self.test_numerical_parser([
            ("Height", []),  # should fail; no values
            ("her height was 1.6m", [1.6]),
            ("Height = 1.23 m", [1.23]),
            ("her height is 1.5m", [1.5]),
            ('''Height 5'8" ''', [m_from_ft_in(feet=5, inches=8)]),
            ("Height 5 ft 8 in", [m_from_ft_in(feet=5, inches=8)]),
            ("Height 5 feet 8 inches", [m_from_ft_in(feet=5, inches=8)]),
        ])
        # *** deal with "tall" and plain "is", e.g.
        # she is 6'2"; she is 1.5m tall


class HeightValidator(ValidatorBase):
    """Validator for Height (see ValidatorBase for explanation)."""
    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(nlpdef=nlpdef,
                         cfgsection=cfgsection,
                         regex_str_list=[Height.HEIGHT],
                         validated_variable=Height.NAME,
                         commit=commit)


# -----------------------------------------------------------------------------
# Weight (mass)
# -----------------------------------------------------------------------------

class Weight(NumericalResultParser):
    """Weight. Handles metric and imperial."""
    METRIC_WEIGHT = r"""
        (                           # capture group
            ( {SIGNED_FLOAT} )          # capture group
            {OPTIONAL_RESULTS_IGNORABLES}
            {KG}
        )
    """.format(SIGNED_FLOAT=SIGNED_FLOAT,
               OPTIONAL_RESULTS_IGNORABLES=OPTIONAL_RESULTS_IGNORABLES,
               KG=KG)
    IMPERIAL_WEIGHT = r"""
        (                           # capture group
            (?:
                ( {SIGNED_FLOAT} )      # capture group
                {OPTIONAL_RESULTS_IGNORABLES}
                {STONES}
                {OPTIONAL_RESULTS_IGNORABLES}
                ( {SIGNED_FLOAT} )      # capture group
                {OPTIONAL_RESULTS_IGNORABLES}
                {LB}
            )
            | (?:
                ( {SIGNED_FLOAT} )      # capture group
                {OPTIONAL_RESULTS_IGNORABLES}
                {STONES}
            )
            | (?:
                ( {SIGNED_FLOAT} )      # capture group
                {OPTIONAL_RESULTS_IGNORABLES}
                {LB}
            )
        )
    """.format(SIGNED_FLOAT=SIGNED_FLOAT,
               OPTIONAL_RESULTS_IGNORABLES=OPTIONAL_RESULTS_IGNORABLES,
               STONES=STONES,
               LB=LB)
    WEIGHT = r"""
        (?:
            \b weigh[ts] \b       # weight, weighs
        )
    """
    REGEX = r"""
        ( {WEIGHT} )                       # group for "BMI" or equivalent
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {TENSE_INDICATOR} )?             # optional group for tense indicator
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {RELATION} )?                    # optional group for relation
        {OPTIONAL_RESULTS_IGNORABLES}
        (?:
            {METRIC_WEIGHT}
            | {IMPERIAL_WEIGHT}
        )
    """.format(
        WEIGHT=WEIGHT,
        OPTIONAL_RESULTS_IGNORABLES=OPTIONAL_RESULTS_IGNORABLES,
        TENSE_INDICATOR=TENSE_INDICATOR,
        RELATION=RELATION,
        METRIC_WEIGHT=METRIC_WEIGHT,
        IMPERIAL_WEIGHT=IMPERIAL_WEIGHT,
        SIGNED_FLOAT=SIGNED_FLOAT,
        KG_PER_SQ_M=KG_PER_SQ_M,
    )

    COMPILED_REGEX = compile_regex(REGEX)
    NAME = "Weight"
    PREFERRED_UNIT_COLUMN = "value_kg"

    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False,
                 debug: bool = False) -> None:
        super().__init__(
            nlpdef=nlpdef,
            cfgsection=cfgsection,
            variable=self.NAME,
            target_unit=self.PREFERRED_UNIT_COLUMN,
            commit=commit
        )
        if debug:
            print("Regex for {}: {}".format(type(self).__name__, self.REGEX))

    def parse(self, text: str,
              debug: bool = False) -> Iterator[Tuple[str, Dict[str, Any]]]:
        for m in self.COMPILED_REGEX.finditer(text):
            if debug:
                print("Match {} for {}".format(m, repr(text)))
            startpos = m.start()
            endpos = m.end()
            matching_text = m.group(0)  # the whole thing
            variable_text = m.group(1)
            tense_indicator = m.group(2)
            relation = m.group(3)
            metric_expression = m.group(4)
            metric_value = m.group(5)
            imperial_expression = m.group(6)
            imperial_st_and_lb_st = m.group(7)
            imperial_st_and_lb_lb = m.group(8)
            imperial_st_only_st = m.group(9)
            imperial_lb_only_lb = m.group(10)

            expression = None
            value_kg = None
            if metric_expression:
                expression = metric_expression
                value_kg = to_float(metric_value)
            elif imperial_expression:
                expression = imperial_expression
                if imperial_st_and_lb_st and imperial_st_and_lb_lb:
                    st = to_float(imperial_st_and_lb_st)
                    lb = to_float(imperial_st_and_lb_lb)
                    value_kg = kg_from_st_lb_oz(stones=st, pounds=lb)
                elif imperial_st_only_st:
                    st = to_float(imperial_st_only_st)
                    value_kg = kg_from_st_lb_oz(stones=st)
                elif imperial_lb_only_lb:
                    lb = to_float(imperial_lb_only_lb)
                    value_kg = kg_from_st_lb_oz(pounds=lb)

            # All left as signed float, as you definitely see things like
            # "weight -0.3 kg" for weight changes.

            tense, relation = common_tense(tense_indicator, relation)

            yield self.tablename, {
                self.FN_VARIABLE_NAME: self.variable,
                self.FN_CONTENT: matching_text,
                self.FN_START: startpos,
                self.FN_END: endpos,
                self.FN_VARIABLE_TEXT: variable_text,
                self.FN_RELATION: relation,
                self.FN_VALUE_TEXT: expression,
                self.target_unit: value_kg,
                self.FN_TENSE: tense,
            }

    def test(self):
        self.test_numerical_parser([
            ("Weight", []),  # should fail; no values
            ("her weight was 60.2kg", [60.2]),
            ("Weight = 52.3kg", [52.3]),
            ("she weighs 61kg", [61]),
            ("she weighs 61 kg", [61]),
            ("she weighs 61 kgs", [61]),
            ("she weighs 61 kilo", [61]),
            ("she weighs 61 kilos", [61]),
            ("she weighs 8 stones ", [kg_from_st_lb_oz(stones=8)]),
            ("she weighs 200 lb", [kg_from_st_lb_oz(pounds=200)]),
            ("she weighs 200 pounds", [kg_from_st_lb_oz(pounds=200)]),
            ("she weighs 6 st 12 lb", [kg_from_st_lb_oz(stones=6, pounds=12)]),
        ])


class WeightValidator(ValidatorBase):
    """Validator for Weight (see ValidatorBase for explanation)."""
    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(nlpdef=nlpdef,
                         cfgsection=cfgsection,
                         regex_str_list=[Weight.WEIGHT],
                         validated_variable=Weight.NAME,
                         commit=commit)


# -----------------------------------------------------------------------------
# Body mass index (BMI)
# -----------------------------------------------------------------------------

class Bmi(SimpleNumericalResultParser):
    """Body mass index (in kg / m^2)."""
    BMI = r"""
        (?:
            {WORD_BOUNDARY}
            (?:
                BMI
                | body \s+ mass \s+ index
            )
            {WORD_BOUNDARY}
        )
    """.format(WORD_BOUNDARY=WORD_BOUNDARY)
    REGEX = r"""
        ( {BMI} )                          # group for "BMI" or equivalent
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {TENSE_INDICATOR} )?             # optional group for tense indicator
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {RELATION} )?                    # optional group for relation
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {SIGNED_FLOAT} )                 # group for value
        {OPTIONAL_RESULTS_IGNORABLES}
        (                                  # group for units
            {KG_PER_SQ_M}
        )?
    """.format(
        BMI=BMI,
        OPTIONAL_RESULTS_IGNORABLES=OPTIONAL_RESULTS_IGNORABLES,
        TENSE_INDICATOR=TENSE_INDICATOR,
        RELATION=RELATION,
        SIGNED_FLOAT=SIGNED_FLOAT,
        KG_PER_SQ_M=KG_PER_SQ_M,
    )
    NAME = "BMI"
    PREFERRED_UNIT_COLUMN = "value_kg_per_sq_m"
    UNIT_MAPPING = {
        KG_PER_SQ_M: 1,       # preferred unit
    }
    # deal with "a BMI of 30"?

    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(
            nlpdef=nlpdef,
            cfgsection=cfgsection,
            regex_str=self.REGEX,
            variable=self.NAME,
            target_unit=self.PREFERRED_UNIT_COLUMN,
            units_to_factor=self.UNIT_MAPPING,
            commit=commit
        )

    def test(self):
        self.test_numerical_parser([
            ("BMI", []),  # should fail; no values
            ("body mass index was 30", [30]),
            ("his BMI (30) is too high", [30]),
            ("BMI 25 kg/sq m", [25]),
            ("BMI was 18.4 kg/m^-2", [18.4]),
            ("ACE 79", []),
        ])


class BmiValidator(ValidatorBase):
    """Validator for Bmi (see ValidatorBase for explanation)."""
    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(nlpdef=nlpdef,
                         cfgsection=cfgsection,
                         regex_str_list=[Bmi.BMI],
                         validated_variable=Bmi.NAME,
                         commit=commit)


# =============================================================================
#  Bedside investigations: BP
# =============================================================================

class Bp(BaseNlpParser):
    """Blood pressure, in mmHg. (Since we produce two variables, SBP and DBP,
    we subclass BaseNlpParser directly.)"""
    BP = r"""
        (?:
            \b blood \s+ pressure \b
            | \b B\.?P\.? \b
        )
    """
    SYSTOLIC_BP = r"""
        (?:
            \b systolic \s+ {BP}
            | \b S\.?B\.?P\.? \b
        )
    """.format(BP=BP)
    DIASTOLIC_BP = r"""
        (?:
            \b diastolic \s+ {BP}
            | \b D\.?B\.?P\.? \b
        )
    """.format(BP=BP)

    TWO_NUMBER_BP = r"""
        ( {SIGNED_FLOAT} )
        (?:
            \b over \b
            | \/
        )
        ( {SIGNED_FLOAT} )
    """.format(SIGNED_FLOAT=SIGNED_FLOAT)
    ONE_NUMBER_BP = SIGNED_FLOAT

    COMPILED_BP = compile_regex(BP)
    COMPILED_SBP = compile_regex(SYSTOLIC_BP)
    COMPILED_DBP = compile_regex(DIASTOLIC_BP)
    COMPILED_ONE_NUMBER_BP = compile_regex(ONE_NUMBER_BP)
    COMPILED_TWO_NUMBER_BP = compile_regex(TWO_NUMBER_BP)
    REGEX = r"""
        (                               # group for "BP" or equivalent
            {SYSTOLIC_BP}               # ... from more to less specific
            | {DIASTOLIC_BP}
            | {BP}
        )
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {TENSE_INDICATOR} )?         # optional group for tense indicator
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {RELATION} )?                # optional group for relation
        {OPTIONAL_RESULTS_IGNORABLES}
        (                              # BP
            {SIGNED_FLOAT}                  # 120
            (?:
                \s*
                (?: \b over \b | \/ )       # /
                \s*
                {SIGNED_FLOAT}              # 80
            )?
        )
        {OPTIONAL_RESULTS_IGNORABLES}
        (                              # group for units
            {MM_HG}
        )?
    """.format(
        BP=BP,
        SYSTOLIC_BP=SYSTOLIC_BP,
        DIASTOLIC_BP=DIASTOLIC_BP,
        OPTIONAL_RESULTS_IGNORABLES=OPTIONAL_RESULTS_IGNORABLES,
        TENSE_INDICATOR=TENSE_INDICATOR,
        RELATION=RELATION,
        SIGNED_FLOAT=SIGNED_FLOAT,
        MM_HG=MM_HG,
    )
    COMPILED_REGEX = compile_regex(REGEX)

    FN_SYSTOLIC_BP_MMHG = 'systolic_bp_mmhg'
    FN_DIASTOLIC_BP_MMHG = 'diastolic_bp_mmhg'

    NAME = "BP"
    UNIT_MAPPING = {
        MM_HG: 1,       # preferred unit
    }

    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(
            nlpdef=nlpdef,
            cfgsection=cfgsection,
            commit=commit
        )
        if nlpdef is None:  # only None for debugging!
            self.tablename = ''
        else:
            self.tablename = nlpdef.opt_str(
                cfgsection, 'desttable', required=True)

    def dest_tables_columns(self) -> Dict[str, List[Column]]:
        return {self.tablename: [
            Column(NumericalResultParser.FN_CONTENT, Text,
                   doc="Matching text contents"),
            Column(NumericalResultParser.FN_START, Integer,
                   doc="Start position (of matching string within whole "
                       "text)"),
            Column(NumericalResultParser.FN_END, Integer,
                   doc="End position (of matching string within whole text)"),
            Column(NumericalResultParser.FN_VARIABLE_TEXT, Text,
                   doc="Text that matched the variable name"),
            Column(NumericalResultParser.FN_RELATION,
                   String(NumericalResultParser.MAX_RELATION_LENGTH),
                   doc="Text that matched the mathematical relationship "
                       "between variable and value (e.g. '=', '<='"),
            Column(NumericalResultParser.FN_VALUE_TEXT,
                   String(NumericalResultParser.MAX_VALUE_TEXT_LENGTH),
                   doc="Matched numerical value, as text"),
            Column(NumericalResultParser.FN_UNITS,
                   String(NumericalResultParser.MAX_UNITS_LENGTH),
                   doc="Matched units, as text"),
            Column(self.FN_SYSTOLIC_BP_MMHG, Float,
                   doc="Systolic BP in mmHg"),
            Column(self.FN_DIASTOLIC_BP_MMHG, Float,
                   doc="Diastolic BP in mmHg"),
            Column(NumericalResultParser.FN_TENSE,
                   String(NumericalResultParser.MAX_TENSE_LENGTH),
                   doc="Tense indicator, if known (e.g. '{}', '{}')".format(
                       PAST, PRESENT)),
        ]}

    def parse(self, text: str,
              debug: bool = False) -> Iterator[Tuple[str, Dict[str, Any]]]:
        for m in self.COMPILED_REGEX.finditer(text):
            if debug:
                print("Match {} for {}".format(m, repr(text)))
            startpos = m.start()
            endpos = m.end()
            matching_text = m.group(0)  # the whole thing
            variable_text = m.group(1)
            tense_indicator = m.group(2)
            relation = m.group(3)
            value_text = m.group(4)
            units = m.group(5)

            sbp = None
            dbp = None
            if self.COMPILED_SBP.match(variable_text):
                if self.COMPILED_ONE_NUMBER_BP.match(value_text):
                    sbp = to_pos_float(value_text)
            elif self.COMPILED_DBP.match(variable_text):
                if self.COMPILED_ONE_NUMBER_BP.match(value_text):
                    dbp = to_pos_float(value_text)
            elif self.COMPILED_BP.match(variable_text):
                bpmatch = self.COMPILED_TWO_NUMBER_BP.match(value_text)
                if bpmatch:
                    sbp = to_pos_float(bpmatch.group(1))
                    dbp = to_pos_float(bpmatch.group(2))
            if sbp is None and dbp is None:
                log.warning("Failed interpretation: {}".format(matching_text))
                continue

            tense, relation = common_tense(tense_indicator, relation)

            yield self.tablename, {
                NumericalResultParser.FN_CONTENT: matching_text,
                NumericalResultParser.FN_START: startpos,
                NumericalResultParser.FN_END: endpos,
                NumericalResultParser.FN_VARIABLE_TEXT: variable_text,
                NumericalResultParser.FN_RELATION: relation,
                NumericalResultParser.FN_VALUE_TEXT: value_text,
                NumericalResultParser.FN_UNITS: units,
                self.FN_SYSTOLIC_BP_MMHG: sbp,
                self.FN_DIASTOLIC_BP_MMHG: dbp,
                NumericalResultParser.FN_TENSE: tense,
            }

    def test_bp_parser(
            self,
            test_expected_list: List[Tuple[str, List[Tuple[float,
                                                           float]]]]) -> None:
        print("Testing parser: {}".format(type(self).__name__))
        for test_string, expected_values in test_expected_list:
            actual_values = list(
                (x[self.FN_SYSTOLIC_BP_MMHG], x[self.FN_DIASTOLIC_BP_MMHG])
                for t, x in self.parse(test_string)
            )
            assert actual_values == expected_values, (
                """Parser {}: Expected {}, got {}, when parsing {}""".format(
                    type(self).__name__,
                    expected_values,
                    actual_values,
                    repr(test_string)
                )
            )
        print("... OK")

    def test(self):
        self.test_bp_parser([
            ("BP", []),  # should fail; no values
            ("his blood pressure was 120/80", [(120, 80)]),
            ("BP 120/80 mmhg", [(120, 80)]),
            ("systolic BP 120", [(120, None)]),
            ("diastolic BP 80", [(None, 80)]),
            ("BP-130/70", [(130, 70)]),
            # Unsure if best to take abs value.
            # One reason not to might be if people express changes, e.g.
            # "BP change -40/-10", but I very much doubt it.
            # Went with abs value using to_pos_float().
        ])


class BpValidator(ValidatorBase):
    """Validator for Bp (see ValidatorBase for explanation)."""
    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(nlpdef=nlpdef,
                         cfgsection=cfgsection,
                         regex_str_list=[Bp.REGEX],
                         validated_variable=Bp.NAME,
                         commit=commit)


# =============================================================================
#  Command-line entry point
# =============================================================================

if __name__ == '__main__':
    height = Height(None, None)
    height.test()
    weight = Weight(None, None)
    weight.test()
    bmi = Bmi(None, None)
    bmi.test()
    bp = Bp(None, None)
    bp.test()



#!/usr/bin/env python
# crate_anon/tools/generate_new_django_secret_key.py

# See django.core.management.commands.startproject.Command.handle

from django.utils.crypto import get_random_string


def main():
    chars = 'abcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*(-_=+)'
    key = get_random_string(50, chars)
    print(key)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding=utf-8
# crate_anon/anonymise/make_demo_database.py

"""
Makes a giant test database for anonymisation testing.

Author: Rudolf Cardinal
Created at: 21 Feb 2015
Last update: 22 Nov 2015

Copyright/licensing:

    Copyright (C) 2015-2016 Rudolf Cardinal (rudolf@pobox.com).
    Department of Psychiatry, University of Cambridge.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

See also:

    http://www.ncbi.nlm.nih.gov/pmc/articles/PMC3751474/
    http://www.ncbi.nlm.nih.gov/pmc/articles/PMC3751474/table/T7/

After anonymisation, check with:

    SELECT * FROM anonymous_output.notes WHERE brcid IN (
        SELECT brcid
        FROM anonymous_mapping.secret_map
        WHERE patient_id < 2
    );
    SELECT * FROM test.patients WHERE patient_id < 2;

"""

import argparse
import datetime
import logging
import os
import random
import subprocess

from sqlalchemy import (
    create_engine,
    BigInteger,
    Boolean,
    Column,
    DateTime,
    ForeignKey,
    Integer,
    LargeBinary,
    MetaData,
    String,
    Text,
)
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, sessionmaker
from sqlalchemy.sql import text

from crate_anon.anonymise.constants import (
    MYSQL_CHARSET,
    MYSQL_TABLE_KWARGS,
)
from crate_anon.common.logsupport import configure_logger_for_colour

log = logging.getLogger(__name__)
metadata = MetaData()
Base = declarative_base(metadata=metadata)

# =============================================================================
# Constants
# =============================================================================

CONSOLE_ENCODING = 'utf8'
REPORT_EVERY = 50
BASE_DOB = datetime.datetime(day=1, month=10, year=1980)
DT_FORMATS = [
    "%d %b %Y",  # e.g. 24 Jul 2013
    "%d %B %Y",  # e.g. 24 July 2013
    "%a %d %B %Y",  # e.g. Wed 24 July 2013
    "%d %B %Y, %H:%M %z",  # ... e.g. 24 July 2013, 20:04 +0100
    "%a %d %B %Y, %H:%M %z",  # ... e.g. Wed 24 July 2013, 20:04 +0100
    "%a %d %B %Y, %H:%M",  # ... e.g. Wed 24 July 2013, 20:04
    "%a %d %b %Y, %H:%M",  # ... e.g. Wed 24 Jul 2013, 20:04
    "%d %B %Y, %H:%M:%S %z",
    "%d %b %Y, %H:%M %z",
    "%d %b %Y, %H:%M:%S %z",
    "%H:%M",
    "%Y-%m-%dT%H:%M:%S%z",  # e.g. 2013-07-24T20:04:07+0100
    "%Y-%m-%d",  # e.g. 2013-07-24
    "%Y-%m-%dT%H%M",  # e.g. 20130724T2004
    "%Y-%m-%d",  # e.g. 20130724
    "%Y%m%d%H%M%S%z",  # e.g. 20130724200407+0100
    "%Y%m%d",  # e.g. 20130724
    "%Y-%m-%dT%H:%M:%SZ",  # e.g. 2013-07-24T20:03:07Z
    "%d/%m/%Y %H:%M",  # e.g. 01/12/2014 09:45
]

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))
DEFAULT_DOCDIR = os.path.join(CURRENT_DIR, os.pardir, os.pardir,
                              "testdocs_for_text_extraction")
DOCTEST_DOC = os.path.join(DEFAULT_DOCDIR, 'doctest.doc')
DOCTEST_DOCX = os.path.join(DEFAULT_DOCDIR, 'doctest.docx')
DOCTEST_ODT = os.path.join(DEFAULT_DOCDIR, 'doctest.odt')
DOCTEST_PDF = os.path.join(DEFAULT_DOCDIR, 'doctest.pdf')

MAX_EXT_LENGTH_WITH_DOT = 10

# =============================================================================
# BLOB type
# =============================================================================

"""
MySQL: http://dev.mysql.com/doc/refman/5.7/en/storage-requirements.html
    TINYBLOB: up to 2^8 bytes
    BLOB: up to 2^16 bytes = 64 KiB
    MEDIUMBLOB: up to 2^24 bytes = 16 MiB  <-- minimum for docs
    LONGBLOB: up to 2^32 bytes = 4 GiB

    VARBINARY: up to 65535 = 64 KiB

SQL Server: https://msdn.microsoft.com/en-us/library/ms188362.aspx
    BINARY: up to 8000 bytes = 8 KB
    VARBINARY(MAX): up to 2^31 - 1 bytes = 2 GiB <-- minimum for docs
    IMAGE: deprecated; up to 2^31 - 1 bytes = 2 GiB
        https://msdn.microsoft.com/en-us/library/ms187993.aspx

SQL Alchemy:
    _Binary: base class
    LargeBinary: translates to BLOB in MySQL
    VARBINARY, as an SQL base data type
    dialects.mysql.base.LONGBLOB
    dialects.mssql.base.VARBINARY

Therefore, we can take the LargeBinary type and modify it:
"""


# http://docs.sqlalchemy.org/en/latest/core/custom_types.html
# noinspection PyUnusedLocal
@compiles(LargeBinary, 'mysql')
def compile_blob_mysql(type_, compiler, **kw):
    return "LONGBLOB"  # would have been "BLOB"


# If this goes wrong for future versions of SQL Server, write another
# specializer to produce "VARBINARY(MAX)" instead of "IMAGE". I haven't done
# that because it may be that SQL Alchemy is reading the SQL Server version
# (it definitely executes "select @@version") and specializing accordingly.


# =============================================================================
# Tables
# =============================================================================

class Patient(Base):
    __tablename__ = 'patient'
    __table_args__ = MYSQL_TABLE_KWARGS

    patient_id = Column(Integer, primary_key=True, autoincrement=False)
    forename = Column(String(50))
    surname = Column(String(50))
    dob = Column(DateTime)
    nullfield = Column(Integer)
    nhsnum = Column(BigInteger)
    phone = Column(String(50))
    postcode = Column(String(50))
    optout = Column(Boolean, default=False)
    related_patient_id = Column(Integer)


class Note(Base):
    __tablename__ = 'note'
    __table_args__ = MYSQL_TABLE_KWARGS

    note_id = Column(Integer, primary_key=True)
    patient_id = Column(Integer, ForeignKey('patient.patient_id'))
    note = Column(Text)

    patient = relationship("Patient")


class BlobDoc(Base):
    __tablename__ = 'blobdoc'
    __table_args__ = MYSQL_TABLE_KWARGS

    blob_doc_id = Column(Integer, primary_key=True)
    patient_id = Column(Integer, ForeignKey('patient.patient_id'))
    blob = Column(LargeBinary)  # modified as above!
    extension = Column(String(MAX_EXT_LENGTH_WITH_DOT))

    patient = relationship("Patient")

    def __init__(self, patient: Patient, filename: str) -> None:
        _, extension = os.path.splitext(filename)
        with open(filename, 'rb') as f:
            contents = f.read()  # will be of type 'bytes'
        super().__init__(patient=patient, blob=contents, extension=extension)


class FilenameDoc(Base):
    __tablename__ = 'filenamedoc'
    __table_args__ = MYSQL_TABLE_KWARGS

    filename_doc_id = Column(Integer, primary_key=True)
    patient_id = Column(Integer, ForeignKey('patient.patient_id'))
    filename = Column(Text)

    patient = relationship("Patient")


# noinspection PyPep8Naming
def main() -> None:
    default_size = 0
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "url",
        help=(
            "SQLAlchemy database URL. Append ?charset=utf8, e.g. "
            "mysql+mysqldb://root:password@127.0.0.1:3306/test?charset=utf8 ."
            " WARNING: If you get the error 'MySQL has gone away', increase "
            "the max_allowed_packet parameter in my.cnf (e.g. to 32M)."))
    parser.add_argument(
        "--size", type=int, default=default_size, choices=[0, 1, 2, 3],
        help="Make tiny (0), small (1), medium (2), or large (3) database "
             "(default={})".format(default_size))
    parser.add_argument('--verbose', '-v', action='store_true',
                        help="Be verbose")
    parser.add_argument("--echo", action="store_true", help="Echo SQL")
    parser.add_argument(
        "--doctest-doc", default=DOCTEST_DOC,
        help="Test file for .DOC (default: {})".format(DOCTEST_DOC))
    parser.add_argument(
        "--doctest-docx", default=DOCTEST_DOCX,
        help="Test file for .DOCX (default: {})".format(DOCTEST_DOCX))
    parser.add_argument(
        "--doctest-odt", default=DOCTEST_ODT,
        help="Test file for .ODT (default: {})".format(DOCTEST_ODT))
    parser.add_argument(
        "--doctest-pdf", default=DOCTEST_PDF,
        help="Test file for .PDF (default: {})".format(DOCTEST_PDF))
    args = parser.parse_args()

    nwords = 10000
    if args.size == 0:
        n_patients = 20
        notes_per_patient = 1
        words_per_note = 100
    elif args.size == 1:
        n_patients = 100
        notes_per_patient = 5
        words_per_note = 100
    elif args.size == 2:
        n_patients = 100
        notes_per_patient = 100
        words_per_note = 1000
    elif args.size == 3:
        # about 1.4 Gb
        n_patients = 1000
        notes_per_patient = 100
        words_per_note = 1000
    else:
        assert False, "Bad size parameter"
    loglevel = logging.DEBUG if args.verbose else logging.INFO
    rootlogger = logging.getLogger()
    configure_logger_for_colour(rootlogger, level=loglevel)

    # 0. Announce intentions

    log.info("n_patients={}, notes_per_patient={}, words_per_note={}".format(
        n_patients, notes_per_patient, words_per_note))

    # 1. Get words

    log.info("Fetching words.")
    words = subprocess.check_output([
        "grep",
        "-v", "'s",
        "-m", str(nwords),
        "/usr/share/dict/words"
    ]).decode(CONSOLE_ENCODING).splitlines()

    # 2. Open database

    log.info("Opening database.")
    log.debug("URL: {}".format(args.url))
    engine = create_engine(args.url, echo=args.echo, encoding=MYSQL_CHARSET)
    session = sessionmaker(bind=engine)()

    # 3. Create tables

    log.info("Creating tables.")
    metadata.drop_all(engine, checkfirst=True)
    metadata.create_all(engine, checkfirst=True)

    # 4. Insert

    log.info("Aiming for a total of {} words in notes.".format(
        n_patients * notes_per_patient * words_per_note))

    log.info("Inserting data.")

    # Special extra patient

    p1 = Patient(
        patient_id=1,
        forename="Ronald Gibbet",
        surname="MacDonald",
        dob=datetime.datetime(day=11, month=11, year=1911),
        nhsnum=123456,
        phone="(01223)-123456",
        postcode="CB2 3EB",
    )
    session.add(p1)
    n1 = Note(
        patient=p1,
        note="""
Ronald MacDonald lived on a farm and kept a gibbet for scaring off
small animals. He was born on 11 Nov 1911 and was very proud of this.
His cat’s name was Flitterwick. It did not like the gibbets.
Ronalds other passion was blimping.
A typo might be RonaldMacDonald.
His phone number was 0122-312-3456, or 01223-123456, or (01223) 123456,
or 01223 123 456, or 01223 123456.
His NHS number was 123.456 or possibly 12 34 56.
His postcode was CB2 3EB, or possible CB23EB, or CB2, or 3EB.

Some HTML encoding is &amp; and &lt;.
An HTML tag is <a href="http://somewhere">this link</a>.
Start aspirin 75mg od. Remains on Lipitor 40mg nocte.
For haloperidol 2mg po prn max qds.
Start amoxicillin 500 mg b.i.d. for 7 days.

Some numerical results:
His CRP is 10. His previous CRP was <13 mg/dl.
Sodium 140.
TSH 3.5; urea normal.
Height 1.82m, weight 75kg, BMI 22.6. BP 135/82.
MMSE 28/30. ACE-R 72, ACE-II 73, ACE 73.
ESR 16 (H) mm/h.
WBC 9.2; neutrophils 4.3; lymphocytes 2.6; eosinophils 0.4; monocytes 1.2;
basophils 0.6.
        """
    )
    session.add(n1)
    for filename in [DOCTEST_DOC, DOCTEST_DOCX, DOCTEST_ODT, DOCTEST_PDF]:
        bd = BlobDoc(patient=p1, filename=filename)
        session.add(bd)
        fd = FilenameDoc(patient=p1, filename=filename)
        session.add(fd)

    p2 = Patient(
        patient_id=2,
        forename="Bob D'Souza",
        surname="",
        dob=datetime.datetime(day=11, month=11, year=1911),
        nhsnum=234567,
        phone="(01223)-234567",
        postcode="CB2 3EB",
        related_patient_id=1,
    )
    session.add(p2)
    n2 = Note(
        patient=p2,
        note="""
Bob D'Souza, also known as Bob, or Mr DSouza, or sometimes Mr D Souza,
or the D'Souza bloke down the road, or BobDSouza or BobD'Souza.
His phone number was 0122-312-3456, or 01223-123456, or (01223) 123456,
or 01223 123 456, or 01223 123456.
His NHS number was 123.456 or possibly 12 34 56 or 123456, perhaps.
His postcode was CB2 3EB, or possible CB23EB, or CB2, or 3EB.
Bob Hope visited Seattle.
Bob took venlafaxine 375 M/R od, and is due to start clozapine 75mg bd.
        """
    )
    session.add(n2)

    # A bunch of patients
    random.seed(1)
    prev_forename = ''
    prev_surname = ''
    for p in range(n_patients):
        if p % REPORT_EVERY == 0:
            log.info("patient {}".format(p))
        forename = words[(p + 1) % nwords] + " " + words[(p + 10) % nwords]
        surname = words[(p + 2) % nwords]
        dob = BASE_DOB + datetime.timedelta(days=p)
        ok_date = dob + datetime.timedelta(days=1)
        nhsnum = random.randint(1, 9999999999)
        patient = Patient(
            patient_id=p + 3,
            forename=forename,
            surname=surname,
            dob=dob,
            nhsnum=nhsnum,
            phone="123456",
            postcode="CB2 3EB",
            related_patient_id=p + 2,  # one back from patient_id
        )
        session.add(patient)
        patient_id = patient.patient_id
        dates = "DATES: " + (
            " ".join([dob.strftime(fmt) for fmt in DT_FORMATS]) +
            " ".join([ok_date.strftime(fmt) for fmt in DT_FORMATS])
        ) + ". "
        fname = "FORENAME: " + forename + ". "
        sname = "SURNAME: " + surname + ". "
        rname = "RELATIVE: " + prev_forename + " " + prev_surname + ". "
        numbers = "NUMBERS: {}, {}, {}. ".format(patient_id, patient_id + 1,
                                                 nhsnum)
        for n in range(notes_per_patient):
            wstr = " ".join(words[p % nwords:(p + words_per_note) % nwords])
            note = Note(
                patient=patient,
                note=fname + sname + rname + numbers + dates + wstr,
            )
            session.add(note)
        prev_forename = forename
        prev_surname = surname

    # 5. Commit

    log.info("Committing...")
    session.commit()
    log.info("Done.")

    # 6. Report size

    if engine.dialect.name == 'mysql':
        log.info("Done. Database size:")
        sql = """
            SELECT
                table_schema,
                table_name,
                table_rows,
                data_length,
                index_length,
                ROUND(((data_length + index_length) / (1024 * 1024)), 2)
                  AS "Size_MB"
            FROM
                information_schema.tables
            WHERE table_schema = DATABASE()
        """
        rows = session.execute(text(sql))
        for r in rows:
            print(
                "schema={}, table={}, rows={}, data_length={}, "
                "index_length={}, size_MB={}".format(*r)
            )


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-07-03 15:30
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('consent', '0003_auto_20160628_1301'),
    ]

    operations = [
        migrations.AddField(
            model_name='dummypatientsourceinfo',
            name='pt_discharge_date',
            field=models.DateField(blank=True, null=True, verbose_name='Patient date of discharge'),  # noqa
        ),
        migrations.AddField(
            model_name='patientlookup',
            name='pt_discharge_date',
            field=models.DateField(blank=True, null=True, verbose_name='Patient date of discharge'),  # noqa
        ),
    ]



#!/usr/bin/env python
# crate_anon/crateweb/consent/forms.py

import logging
from django import forms
from django.conf import settings
from django.db.models import Q
from crate_anon.crateweb.extra.forms import (
    MultipleNhsNumberAreaField,
    MultipleWordAreaField,
    SingleNhsNumberField,
)
from crate_anon.crateweb.consent.models import (
    ClinicianResponse,
    Study,
)

log = logging.getLogger(__name__)


class SingleNhsNumberForm(forms.Form):
    nhs_number = SingleNhsNumberField(label="NHS number")


def get_queryset_possible_contact_studies():
    return (
        Study.objects
        .filter(patient_contact=True)
        .filter(approved_by_rec=True)
        .filter(approved_locally=True)
        .exclude(study_details_pdf='')
        .exclude(lead_researcher__profile__title='')
        .exclude(lead_researcher__first_name='')
        .exclude(lead_researcher__last_name='')
    )


class AbstractContactRequestForm(forms.Form):
    def clean(self):
        cleaned_data = super().clean()

        study = cleaned_data.get("study")
        if not study:
            raise forms.ValidationError("Must specify study")

        request_direct_approach = cleaned_data.get("request_direct_approach")
        if request_direct_approach and not study.request_direct_approach:
            raise forms.ValidationError(
                "Study not approved for direct approach.")


class SuperuserSubmitContactRequestForm(AbstractContactRequestForm):
    study = forms.ModelChoiceField(
        queryset=get_queryset_possible_contact_studies())
    request_direct_approach = forms.BooleanField(
        label="Request direct approach to patient, if available "
              "(UNTICK to ask clinician for additional info)",
        required=False,
        initial=True)
    nhs_numbers = MultipleNhsNumberAreaField(label='NHS numbers',
                                             required=False)
    rids = MultipleWordAreaField(
        label='{} (RID)'.format(settings.SECRET_MAP['RID_FIELD']),
        required=False)
    mrids = MultipleWordAreaField(
        label='{} (MRID)'.format(settings.SECRET_MAP['MASTER_RID_FIELD']),
        required=False)


class ResearcherSubmitContactRequestForm(AbstractContactRequestForm):
    study = forms.ModelChoiceField(queryset=Study.objects.all())
    request_direct_approach = forms.BooleanField(
        label="Request direct approach to patient, if available "
              "(UNTICK to ask clinician for additional info)",
        required=False,
        initial=True)
    rids = MultipleWordAreaField(
        label='{} (RID)'.format(settings.SECRET_MAP['RID_FIELD']),
        required=False)
    mrids = MultipleWordAreaField(
        label='{} (MRID)'.format(settings.SECRET_MAP['MASTER_RID_FIELD']),
        required=False)

    def __init__(self, user, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['study'].queryset = (
            get_queryset_possible_contact_studies()
            .filter(Q(lead_researcher=user) | Q(researchers__in=[user]))
            .distinct()
        )
        # https://docs.djangoproject.com/en/1.8/ref/models/querysets/#field-lookups  # noqa
        # http://stackoverflow.com/questions/5329586/django-modelchoicefield-filtering-query-set-and-setting-default-value-as-an-obj  # noqa


class ClinicianResponseForm(forms.ModelForm):
    class Meta:
        model = ClinicianResponse
        fields = [
            'token',
            'email_choice',
            'response',
            'veto_reason',
            'ineligible_reason',
            'pt_uncontactable_reason',
            'clinician_confirm_name',
        ]
        widgets = {
            'token': forms.HiddenInput(),
            'email_choice': forms.HiddenInput(),
        }



#!/usr/bin/env python
# crate_anon/nlp_manager/parse_cognitive.py

import logging
from typing import Optional

from crate_anon.nlp_manager.nlp_definition import NlpDefinition
from crate_anon.nlp_manager.regex_parser import (
    OPTIONAL_RESULTS_IGNORABLES,
    RELATION,
    SimpleNumericalResultParser,
    TENSE_INDICATOR,
    ValidatorBase,
    WORD_BOUNDARY,
)
from crate_anon.nlp_manager.regex_numbers import SIGNED_FLOAT
from crate_anon.nlp_manager.regex_units import out_of, out_of_anything, SCORE

log = logging.getLogger(__name__)


# =============================================================================
# Mini-mental state examination (MMSE)
# =============================================================================

class Mmse(SimpleNumericalResultParser):
    """Mini-mental state examination (MMSE)."""
    MMSE = r"""
        (?:
            {WORD_BOUNDARY}
            (?:
                MMSE
                | mini[-\s]*mental (?: \s+ state)? (?: \s+ exam(?:ination)? )?
            )
            {WORD_BOUNDARY}
        )
    """.format(WORD_BOUNDARY=WORD_BOUNDARY)
    OUT_OF_30 = out_of(30)
    REGEX = r"""
        ( {MMSE} )                         # group for "MMSE" or equivalent
        {OPTIONAL_RESULTS_IGNORABLES}
        {SCORE}?                           # optional "score" or similar
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {TENSE_INDICATOR} )?             # optional group for tense indicator
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {RELATION} )?                    # optional group for relation
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {SIGNED_FLOAT} )                 # group for value
        {OPTIONAL_RESULTS_IGNORABLES}
        (                                  # group for units
            {OUT_OF_30}                         # good
            | {OUT_OF_ANYTHING}                 # bad
        )?
    """.format(
        MMSE=MMSE,
        OPTIONAL_RESULTS_IGNORABLES=OPTIONAL_RESULTS_IGNORABLES,
        SCORE=SCORE,
        TENSE_INDICATOR=TENSE_INDICATOR,
        RELATION=RELATION,
        SIGNED_FLOAT=SIGNED_FLOAT,
        OUT_OF_30=OUT_OF_30,
        OUT_OF_ANYTHING=out_of_anything(),
    )
    NAME = "MMSE"
    PREFERRED_UNIT_COLUMN = "out_of_30"
    UNIT_MAPPING = {
        OUT_OF_30: 1,       # preferred unit
        # not out_of_anything()
    }

    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(
            nlpdef=nlpdef,
            cfgsection=cfgsection,
            regex_str=self.REGEX,
            variable=self.NAME,
            target_unit=self.PREFERRED_UNIT_COLUMN,
            units_to_factor=self.UNIT_MAPPING,
            commit=commit
        )

    def test(self):
        self.test_numerical_parser([
            ("MMSE", []),  # should fail; no values
            ("MMSE 30/30", [30]),
            ("MMSE 25 / 30", [25]),
            ("MMSE 25 / 29", []),
            ("MMSE 25 / 31", []),
            ("mini-mental state exam 30", [30]),
            ("minimental 25", [25]),
            ("MMSE 30", [30]),
            ("MMSE score was 30", [30]),
            ("ACE 79", []),
        ])


class MmseValidator(ValidatorBase):
    """Validator for Mmse (see ValidatorBase for explanation)."""
    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(nlpdef=nlpdef,
                         cfgsection=cfgsection,
                         regex_str_list=[Mmse.MMSE],
                         validated_variable=Mmse.NAME,
                         commit=commit)


# =============================================================================
# Addenbrooke's Cognitive Examination (ACE, ACE-R, ACE-III)
# =============================================================================

class Ace(SimpleNumericalResultParser):
    """Addenbrooke's Cognitive Examination (ACE, ACE-R, ACE-III)."""
    ACE = r"""
        (?:
            {WORD_BOUNDARY}
            (?:
                ACE
                | (?:
                    Addenbrooke'?s \s+ cognitive \s+
                    (?: (?:evaluation) | exam(?:ination)? )
                )
            )
            (?:
                (?: -? 3 )
                | (?: \s* -? \s* (?: R | III ) )
            )?
            {WORD_BOUNDARY}
        )
    """.format(WORD_BOUNDARY=WORD_BOUNDARY)
    OUT_OF_100 = out_of(100)
    REGEX = r"""
        ( {ACE} )                          # group for "ACE" or equivalent
        {OPTIONAL_RESULTS_IGNORABLES}
        {SCORE}?                           # optional "score" or similar
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {TENSE_INDICATOR} )?             # optional group for tense indicator
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {RELATION} )?                    # optional group for relation
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {SIGNED_FLOAT} )                 # group for value
        {OPTIONAL_RESULTS_IGNORABLES}
        (                                  # group for units
            {OUT_OF_100}                        # good
            | {OUT_OF_ANYTHING}                 # bad
        )?
    """.format(
        ACE=ACE,
        OPTIONAL_RESULTS_IGNORABLES=OPTIONAL_RESULTS_IGNORABLES,
        SCORE=SCORE,
        TENSE_INDICATOR=TENSE_INDICATOR,
        RELATION=RELATION,
        SIGNED_FLOAT=SIGNED_FLOAT,
        OUT_OF_100=OUT_OF_100,
        OUT_OF_ANYTHING=out_of_anything(),
    )
    NAME = "ACE"
    PREFERRED_UNIT_COLUMN = "out_of_100"
    UNIT_MAPPING = {
        OUT_OF_100: 1,       # preferred unit
        # not out_of_anything()
    }

    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(
            nlpdef=nlpdef,
            cfgsection=cfgsection,
            regex_str=self.REGEX,
            variable=self.NAME,
            target_unit=self.PREFERRED_UNIT_COLUMN,
            units_to_factor=self.UNIT_MAPPING,
            commit=commit
        )

    def test(self):
        self.test_numerical_parser([
            ("MMSE", []),
            ("MMSE 30/30", []),
            ("MMSE 25 / 30", []),
            ("mini-mental state exam 30", []),
            ("minimental 25", []),
            ("MMSE 30", []),
            ("ACE 79", [79]),
            ("ACE 79/100", [79]),
            ("ACE 79/95", []),
            ("ACE 79 / 100", [79]),
            ("Addenbrooke's cognitive examination 79", [79]),
            ("Addenbrookes cognitive evaluation 79", [79]),
            ("ACE-R 79", [79]),
            ("ACE-R 79 out of 100", [79]),
            ("ACE-III 79", [79]),
            ("ACE-III score was 79", [79]),
            ("ACE-3 79", [79]),
            ("ACE R 79", [79]),
            ("ACE III 79", [79]),
            ("ACE 3 79", [3]),  # nasty; not easy to cope with this well.
            ("ACE 3", [3]),
        ])


class AceValidator(ValidatorBase):
    """Validator for Ace (see ValidatorBase for explanation)."""
    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(nlpdef=nlpdef,
                         cfgsection=cfgsection,
                         regex_str_list=[Ace.ACE],
                         validated_variable=Ace.NAME,
                         commit=commit)


# =============================================================================
# Mini-Addenbrooke's Cognitive Examination (M-ACE)
# =============================================================================

class MiniAce(SimpleNumericalResultParser):
    """Mini-Addenbrooke's Cognitive Examination (M-ACE)."""
    MACE = r"""
        (?:
            {WORD_BOUNDARY}
            (?:
                mini | M
            )
            \s* -? \s*
            (?:
                ACE
                | (?:
                    Addenbrooke'?s \s+ cognitive \s+
                    (?: (?:evaluation) | exam(?:ination)? )
                )
            )
            {WORD_BOUNDARY}
        )
    """.format(WORD_BOUNDARY=WORD_BOUNDARY)
    OUT_OF_30 = out_of(30)
    REGEX = r"""
        ( {MACE} )                         # group for "M-ACE" or equivalent
        {OPTIONAL_RESULTS_IGNORABLES}
        {SCORE}?                           # optional "score" or similar
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {TENSE_INDICATOR} )?             # optional group for tense indicator
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {RELATION} )?                    # optional group for relation
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {SIGNED_FLOAT} )                 # group for value
        {OPTIONAL_RESULTS_IGNORABLES}
        (                                  # group for units
            {OUT_OF_30}                         # good
            | {OUT_OF_ANYTHING}                 # bad
        )?
    """.format(
        MACE=MACE,
        OPTIONAL_RESULTS_IGNORABLES=OPTIONAL_RESULTS_IGNORABLES,
        SCORE=SCORE,
        TENSE_INDICATOR=TENSE_INDICATOR,
        RELATION=RELATION,
        SIGNED_FLOAT=SIGNED_FLOAT,
        OUT_OF_30=OUT_OF_30,
        OUT_OF_ANYTHING=out_of_anything(),
    )
    NAME = "MiniACE"
    PREFERRED_UNIT_COLUMN = "out_of_30"
    UNIT_MAPPING = {
        OUT_OF_30: 1,       # preferred unit
        # not out_of_anything()
    }

    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(
            nlpdef=nlpdef,
            cfgsection=cfgsection,
            regex_str=self.REGEX,
            variable=self.NAME,
            target_unit=self.PREFERRED_UNIT_COLUMN,
            units_to_factor=self.UNIT_MAPPING,
            commit=commit
        )

    def test(self):
        self.test_numerical_parser([
            ("MMSE 30", []),
            ("ACE 79", []),
            ("ACE 79/100", []),
            ("Addenbrooke's cognitive examination 79", []),
            ("Addenbrookes cognitive evaluation 79", []),
            ("mini-Addenbrooke's cognitive examination 79", [79]),
            ("mini-Addenbrookes cognitive evaluation 79", [79]),
            ("M-ACE 20", [20]),
            ("M-ACE score is 20", [20]),
            ("M-ACE 29/30", [29]),
            ("M-ACE 29/29", []),
            ("MACE 29", [29]),
            ("mini-ACE 29", [29]),
        ])


class MiniAceValidator(ValidatorBase):
    """Validator for MiniAce (see ValidatorBase for explanation)."""
    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(nlpdef=nlpdef,
                         cfgsection=cfgsection,
                         regex_str_list=[MiniAce.MACE],
                         validated_variable=MiniAce.NAME,
                         commit=commit)


# =============================================================================
# Montreal Cognitive Assessment (MOCA)
# =============================================================================

class Moca(SimpleNumericalResultParser):
    """Montreal Cognitive Assessment (MOCA)."""
    MOCA = r"""
        (?:
            {WORD_BOUNDARY}
            (?:
                MOCA
                | (?:
                    Montreal \s+ cognitive \s+ assessment
                )
            )
            {WORD_BOUNDARY}
        )
    """.format(WORD_BOUNDARY=WORD_BOUNDARY)
    OUT_OF_30 = out_of(30)
    REGEX = r"""
        ( {MOCA} )                         # group for "M-ACE" or equivalent
        {OPTIONAL_RESULTS_IGNORABLES}
        {SCORE}?                           # optional "score" or similar
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {TENSE_INDICATOR} )?             # optional group for tense indicator
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {RELATION} )?                    # optional group for relation
        {OPTIONAL_RESULTS_IGNORABLES}
        ( {SIGNED_FLOAT} )                 # group for value
        {OPTIONAL_RESULTS_IGNORABLES}
        (                                  # group for units
            {OUT_OF_30}                         # good
            | {OUT_OF_ANYTHING}                 # bad
        )?
    """.format(
        MOCA=MOCA,
        OPTIONAL_RESULTS_IGNORABLES=OPTIONAL_RESULTS_IGNORABLES,
        SCORE=SCORE,
        TENSE_INDICATOR=TENSE_INDICATOR,
        RELATION=RELATION,
        SIGNED_FLOAT=SIGNED_FLOAT,
        OUT_OF_30=OUT_OF_30,
        OUT_OF_ANYTHING=out_of_anything(),
    )
    NAME = "MOCA"
    PREFERRED_UNIT_COLUMN = "out_of_30"
    UNIT_MAPPING = {
        OUT_OF_30: 1,       # preferred unit
        # not out_of_anything()
    }

    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(
            nlpdef=nlpdef,
            cfgsection=cfgsection,
            regex_str=self.REGEX,
            variable=self.NAME,
            target_unit=self.PREFERRED_UNIT_COLUMN,
            units_to_factor=self.UNIT_MAPPING,
            commit=commit
        )

    def test(self):
        self.test_numerical_parser([
            ("MOCA 30", [30]),
            ("MOCA 30/30", [30]),
            ("MOCA 25/30", [25]),
            ("MOCA score was 25", [25]),
            ("MOCA 25/29", []),
            ("Montreal Cognitive Assessment 25/30", [25]),
        ])


class MocaValidator(ValidatorBase):
    """Validator for MiniAce (see ValidatorBase for explanation)."""
    def __init__(self,
                 nlpdef: Optional[NlpDefinition],
                 cfgsection: Optional[str],
                 commit: bool = False) -> None:
        super().__init__(nlpdef=nlpdef,
                         cfgsection=cfgsection,
                         regex_str_list=[MiniAce.MACE],
                         validated_variable=MiniAce.NAME,
                         commit=commit)


# =============================================================================
#  Command-line entry point
# =============================================================================

if __name__ == '__main__':
    mmse = Mmse(None, None)
    mmse.test()
    ace = Ace(None, None)
    ace.test()
    mace = MiniAce(None, None)
    mace.test()
    moca = Moca(None, None)
    moca.test()

# support also "scored X on the MOCA"?



#!/usr/bin/env python
# crate_anon/crateweb/extra/admin.py

import logging
from typing import Any, Callable, Dict, List, Type

from django.contrib.admin import AdminSite, ModelAdmin
from django.contrib.admin.views.main import ChangeList
from django.core.urlresolvers import reverse
from django.db.models import Model
from django.forms import ModelForm
from django.http import HttpResponse
from django.http.request import HttpRequest
from django.utils.encoding import force_text
from django.utils.html import escape
from django.utils.translation import ugettext

log = logging.getLogger(__name__)


# =============================================================================
# Action-restricted ModelAdmin classes
# =============================================================================

class ReadOnlyChangeList(ChangeList):
    # See ChangeList in django.contrib.admin.views.main
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if self.is_popup:
            title = ugettext('Select %s')
        else:
            title = ugettext('Select %s to view')
        self.title = title % force_text(self.opts.verbose_name)


class ReadOnlyModelAdmin(ModelAdmin):
    """
    Allows view ("change"), but not add/edit/delete.

    You also need to do this:
        my_admin_site.index_template = 'admin/viewchange_admin_index.html'
    ... to give a modified admin/index.html that says "View/change" not
    "Change".

    """
    # http://stackoverflow.com/questions/3068843/permission-to-view-but-not-to-change-django  # noqa
    # See also http://stackoverflow.com/questions/6680631/django-admin-separate-read-only-view-and-change-view  # noqa
    # django/contrib/admin/templates/admin/change_form.html
    # django/contrib/admin/templatetags/admin_modify.py
    # https://docs.djangoproject.com/en/1.8/ref/contrib/admin/#django.contrib.ModelAdmin.change_view  # noqa

    # Remove the tickbox for deletion, and the top/bottom action bars:
    actions = None

    # When you drill down into a single object, use a custom template
    # that removes the 'save' buttons:
    change_form_template = 'admin/readonly_view_form.html'

    def has_add_permission(self, request: HttpRequest, obj=None) -> bool:
        # Don't let the user add objects.
        return False

    def has_delete_permission(self, request: HttpRequest, obj=None) -> bool:
        # Don't let the user delete objects.
        return False

    # Don't remove has_change_permission, or you won't see anything.
    # def has_change_permission(self, request, obj=None):
    #     return False

    def save_model(self, request: HttpRequest, obj, 
                   form: ModelForm, change: bool):
        # Return nothing to make sure user can't update any data
        pass

    # Make list say "Select [model] to view" not "... change"
    def get_changelist(self, request: HttpRequest, **kwargs) \
            -> Type[ChangeList]:
        return ReadOnlyChangeList

    # Make single object view say "View [model]", not "Change [model]"
    def change_view(self, 
                    request: HttpRequest, 
                    object_id: int, 
                    form_url: str = '', 
                    extra_context: Dict[str, Any] = None) -> HttpResponse:
        extra_context = extra_context or {}
        # noinspection PyProtectedMember
        extra_context["title"] = "View %s" % force_text(
            self.model._meta.verbose_name)
        return super().change_view(request, object_id, form_url,
                                   extra_context=extra_context)


class AddOnlyModelAdmin(ModelAdmin):
    """
    Allows add, but not edit or delete.
    Optional extra class attribute:
        fields_for_viewing
    """
    actions = None

    # When you drill down into a single object, use a custom template
    # that removes the 'save' buttons:
    change_form_template = 'admin/readonly_view_form.html'

    # But keep the default for adding:
    add_form_template = 'admin/change_form.html'

    def has_delete_permission(self, request: HttpRequest, obj=None) -> bool:
        return False

    def get_changelist(self, request: HttpRequest, **kwargs) \
            -> Type[ChangeList]:
        return ReadOnlyChangeList

    # This is an add-but-not-edit class.
    # http://stackoverflow.com/questions/7860612/django-admin-make-field-editable-in-add-but-not-edit  # noqa
    def get_readonly_fields(self, request: HttpRequest, obj=None) -> List[str]:
        if obj:  # obj is not None, so this is an edit
            # self.__class__ is the derived class
            if hasattr(self.__class__, 'fields_for_viewing'):
                return self.__class__.fields_for_viewing
            elif hasattr(self.__class__, 'readonly_fields'):
                return self.__class__.readonly_fields
            else:
                return self.__class__.fields
        else:  # This is an addition
            return []

    def get_fields(self, request: HttpRequest, obj=None) -> List[str]:
        if obj:  # edit (view)
            return self.__class__.fields_for_viewing
        else:
            return self.__class__.fields

    # Make single object view say "View [model]", not "Change [model]"
    def change_view(self, 
                    request: HttpRequest, 
                    object_id: int, 
                    form_url: str = '', 
                    extra_context: Dict[str, Any] = None) -> HttpResponse:
        extra_context = extra_context or {}
        # noinspection PyProtectedMember
        extra_context["title"] = "View %s" % force_text(
            self.model._meta.verbose_name)
        return super().change_view(request, object_id, form_url,
                                   extra_context=extra_context)


class EditOnlyModelAdmin(ModelAdmin):
    """
    Allows editing, but not add or delete.
    Designed for e.g. when you have a fixed set of PKs.
    In that situation, ensure the PK field is in readonly_fields.
    """
    actions = None

    def has_add_permission(self, request: HttpRequest, obj=None) -> bool:
        return False

    def has_delete_permission(self, request: HttpRequest, obj=None) -> bool:
        return False


class EditOnceOnlyModelAdmin(ModelAdmin):
    """
    Allows editing, but not add or delete.
    Designed for e.g. when you have a fixed set of PKs.
    In that situation, ensure the PK field is in readonly_fields.
    """
    actions = None

    change_form_template = 'admin/edit_once_view_form.html'

    def has_add_permission(self, request: HttpRequest, obj=None) -> bool:
        return False

    def has_delete_permission(self, request: HttpRequest, obj=None) -> bool:
        return False


class AllStaffReadOnlyModelAdmin(ReadOnlyModelAdmin):
    """
    ReadOnlyModelAdmin but allows access to all staff, not just superusers.
    (No easy way to make this work via multiple inheritance.)
    """
    def has_module_permission(self, request: HttpRequest) -> bool:
        return request.user.is_staff

    def has_change_permission(self, request: HttpRequest, obj=None) -> bool:
        return request.user.is_staff


# =============================================================================
# Links in model admin
# =============================================================================

# noinspection PyProtectedMember
def admin_view_url(admin_site: AdminSite, 
                   obj, 
                   view_type: str = "change", 
                   current_app: str = None) -> str:
    app_name = obj._meta.app_label.lower()
    model_name = obj._meta.object_name.lower()
    pk = obj.pk
    viewname = "admin:{}_{}_{}".format(app_name, model_name, view_type)
    if current_app is None:
        current_app = admin_site.name
    url = reverse(viewname, args=[pk], current_app=current_app)
    return url


# noinspection PyProtectedMember
def admin_view_fk_link(modeladmin: ModelAdmin,
                       obj,
                       fkfield: str,
                       missing: str = "(None)",
                       use_str: bool = True,
                       view_type: str = "change",
                       current_app: str = None) -> str:
    if not hasattr(obj, fkfield):
        return missing
    linked_obj = getattr(obj, fkfield)
    app_name = linked_obj._meta.app_label.lower()
    model_name = linked_obj._meta.object_name.lower()
    viewname = "admin:{}_{}_{}".format(app_name, model_name, view_type)
    # https://docs.djangoproject.com/en/dev/ref/contrib/admin/#reversing-admin-urls  # noqa
    if current_app is None:
        current_app = modeladmin.admin_site.name
        # ... plus a bit of home-grown magic; see Django source
    url = reverse(viewname, args=[linked_obj.pk], current_app=current_app)
    if use_str:
        label = escape(str(linked_obj))
    else:
        label = "{} {}".format(escape(linked_obj._meta.object_name),
                               linked_obj.pk)
    return '<a href="{}">{}</a>'.format(url, label)


# noinspection PyProtectedMember
def admin_view_reverse_fk_links(modeladmin: ModelAdmin,
                                obj,
                                reverse_fk_set_field: str,
                                missing: str = "(None)",
                                use_str: bool = True,
                                separator: str = "<br>",
                                view_type: str = "change",
                                current_app: str = None) -> str:
    if not hasattr(obj, reverse_fk_set_field):
        return missing
    linked_objs = getattr(obj, reverse_fk_set_field).all()
    if not linked_objs:
        return missing
    first = linked_objs[0]
    app_name = first._meta.app_label.lower()
    model_name = first._meta.object_name.lower()
    viewname = "admin:{}_{}_{}".format(app_name, model_name, view_type)
    if current_app is None:
        current_app = modeladmin.admin_site.name
    links = []
    for linked_obj in linked_objs:
        # log.debug("linked_obj: {}".format(linked_obj))
        url = reverse(viewname, args=[linked_obj.pk], current_app=current_app)

        if use_str:
            label = escape(str(linked_obj))
        else:
            label = "{} {}".format(escape(linked_obj._meta.object_name),
                                   linked_obj.pk)
        links.append('<a href="{}">{}</a>'.format(url, label))
    # log.debug("links: {}".format(links))
    return separator.join(links)


# =============================================================================
# Disable boolean icons for a ModelAdmin field
# =============================================================================
# http://stackoverflow.com/questions/13990846/disable-on-off-icon-for-boolean-field-in-django  # noqa
# ... extended to use closures

def disable_bool_icon(fieldname: str, model: Model) -> Callable[[Any], bool]:
    # noinspection PyUnusedLocal
    def func(self, obj):
        return getattr(obj, fieldname)
    func.boolean = False
    func.admin_order_field = fieldname
    # func.short_description = \
    #     model._meta.get_field_by_name(fieldname)[0].verbose_name
    # get_field_by_name() deprecated in Django 1.9 and will go in 1.10
    # https://docs.djangoproject.com/en/1.8/ref/models/meta/
    # noinspection PyProtectedMember
    func.short_description = \
        model._meta.get_field(fieldname).verbose_name
    return func



#!/usr/bin/env python
# crate_anon/crateweb/research/sql_grammar_mysql.py

"""
Two approaches:

- sqlparse
  https://sqlparse.readthedocs.io/en/latest/
  ... LESS GOOD.

- pyparsing
  https://pyparsing.wikispaces.com/file/view/select_parser.py
  http://pyparsing.wikispaces.com/file/view/simpleSQL.py
  http://sacharya.com/parsing-sql-with-pyparsing/
  http://bazaar.launchpad.net/~eleybourn/temporal-proxy/devel/view/head:/parser.py  # noqa
  http://pyparsing.wikispaces.com/file/view/select_parser.py/158651233/select_parser.py  # noqa

  - Maximum recursion depth exceeded:
    http://pyparsing.wikispaces.com/share/view/11262229

  - Match-first (|) versus match-longest (^):
    http://pyparsing.wikispaces.com/HowToUsePyparsing

  - How to do arithmetic recursive parsing:
    http://stackoverflow.com/questions/1345039

  - Somebody else's MySQL parser:
    https://github.com/gburns/mysql_pyparsing/blob/master/src/sqlparser.py

  - ... or generic SQL:
    http://pyparsing.wikispaces.com/file/view/select_parser.py/158651233/select_parser.py  # noqa
"""

# import itertools
import logging
import re
# import sys

from pyparsing import (
    alphanums,
    # alphas,
    # alphas8bit,
    # CaselessKeyword,
    # CaselessLiteral,
    # CharsNotIn,
    Combine,
    # commaSeparatedList,
    cStyleComment,
    # downcaseTokens,
    delimitedList,
    Forward,
    Group,
    # hexnums,
    infixNotation,
    # Keyword,
    Literal,
    NotAny,
    # nums,
    oneOf,
    OneOrMore,
    opAssoc,
    Optional,
    # Or,
    ParseException,
    ParserElement,
    ParseResults,
    QuotedString,
    Regex,
    # sglQuotedString,
    # srange,
    Suppress,
    Token,
    restOfLine,
    Word,
    ZeroOrMore,
)
import sqlparse

from crate_anon.common.logsupport import main_only_quicksetup_rootlogger

log = logging.getLogger(__name__)


# =============================================================================
# pyparsing helpers
# =============================================================================

def delim_list(expr_: Token,
               delim: str = ",",
               combine: bool = False,
               combine_adjacent: bool = False,
               suppress_delim: bool = False):
    # A better version of delimitedList
    # BUT SEE ALSO ALTERNATIVE: http://stackoverflow.com/questions/37926516
    name = str(expr_) + " [" + str(delim) + " " + str(expr_) + "]..."
    if combine:
        return Combine(expr_ + ZeroOrMore(delim + expr_),
                       adjacent=combine_adjacent).setName(name)
    else:
        if suppress_delim:
            return (expr_ + ZeroOrMore(Suppress(delim) + expr_)).setName(name)
        else:
            return (expr_ + ZeroOrMore(delim + expr_)).setName(name)


WORD_BOUNDARY = r"\b"
# The meaning of \b:
# http://stackoverflow.com/questions/4213800/is-there-something-like-a-counter-variable-in-regular-expression-replace/4214173#4214173  # noqa


def word_regex_element(word: str) -> str:
    return WORD_BOUNDARY + word + WORD_BOUNDARY


def multiple_words_regex_element(words_as_string: str) -> str:
    wordlist = words_as_string.split()
    return WORD_BOUNDARY + "(" + "|".join(wordlist) + ")" + WORD_BOUNDARY


def make_pyparsing_regex(regex_str: str,
                         caseless: bool = False,
                         name: str = None) -> Regex:
    flags = re.IGNORECASE if caseless else 0
    result = Regex(regex_str, flags=flags)
    if name:
        result.setName(name)
    return result


def make_words_regex(words_as_string: str,
                     caseless: bool = False,
                     name: str = None,
                     debug: bool = False) -> Regex:
    regex_str = multiple_words_regex_element(words_as_string)
    if debug:
        log.debug(regex_str)
    return make_pyparsing_regex(regex_str, caseless=caseless, name=name)


def make_regex_except_words(word_pattern: str,
                            negative_words_str: str,
                            caseless: bool = False,
                            name: str = None,
                            debug: bool = False) -> Regex:
    # http://regexr.com/
    regex_str = r"(?!{negative}){positive}".format(
        negative=multiple_words_regex_element(negative_words_str),
        positive=(WORD_BOUNDARY + word_pattern + WORD_BOUNDARY),
    )
    if debug:
        log.debug(regex_str)
    return make_pyparsing_regex(regex_str, caseless=caseless, name=name)


def sql_keyword(word: str) -> Regex:
    regex_str = word_regex_element(word)
    return make_pyparsing_regex(regex_str, caseless=True, name=word)


def bracket(fragment: str) -> str:
    return "(" + fragment + ")"


def single_quote(fragment: str) -> str:
    return "'" + fragment + "'"


# =============================================================================
# MySQL 5.7 grammar in pyparsing
# =============================================================================
# http://dev.mysql.com/doc/refman/5.7/en/dynindex-statement.html

# Keywords, using regexed for speed
AGAINST = sql_keyword("AGAINST")
ALL = sql_keyword("ALL")
ALTER = sql_keyword("ALTER")
AND = sql_keyword("AND")
ANY = sql_keyword("ANY")
AS = sql_keyword("AS")
ASC = sql_keyword("ASC")
BETWEEN = sql_keyword("BETWEEN")
BINARY = sql_keyword("BINARY")
BOOLEAN = sql_keyword("BOOLEAN")
BY = sql_keyword("BY")
CASE = sql_keyword("CASE")
CASCADE = sql_keyword("CASCADE")
COLLATE = sql_keyword("COLLATE")
CREATE = sql_keyword("CREATE")
CROSS = sql_keyword("CROSS")
DATE = sql_keyword("DATE")
DATETIME = sql_keyword("DATETIME")
DELETE = sql_keyword("DELETE")
DEFAULT = sql_keyword("DEFAULT")
DESC = sql_keyword("DESC")
DISTINCT = sql_keyword("DISTINCT")
DISTINCTROW = sql_keyword("DISTINCTROW")
DIV = sql_keyword("DIV")
DROP = sql_keyword("DROP")
ELSE = sql_keyword("ELSE")
END = sql_keyword("END")
ESCAPE = sql_keyword("ESCAPE")
EXISTS = sql_keyword("EXISTS")
EXPANSION = sql_keyword("EXPANSION")
FALSE = sql_keyword("FALSE")
FIRST = sql_keyword("FIRST")
FOR = sql_keyword("FOR")
FORCE = sql_keyword("FORCE")
FROM = sql_keyword("FROM")
GROUP = sql_keyword("GROUP")
HAVING = sql_keyword("HAVING")
HIGH_PRIORITY = sql_keyword("HIGH_PRIORITY")
IGNORE = sql_keyword("IGNORE")
IN = sql_keyword("IN")
INDEX = sql_keyword("INDEX")
INNER = sql_keyword("INNER")
INSERT = sql_keyword("INSERT")
INTERVAL = sql_keyword("INTERVAL")
INTO = sql_keyword("INTO")
IS = sql_keyword("IS")
JOIN = sql_keyword("JOIN")
KEY = sql_keyword("KEY")
LANGUAGE = sql_keyword("LANGUAGE")
LAST = sql_keyword("LAST")
LEFT = sql_keyword("LEFT")
LIKE = sql_keyword("LIKE")
LIMIT = sql_keyword("LIMIT")
MATCH = sql_keyword("MATCH")
MAX_STATEMENT_TIME = sql_keyword("MAX_STATEMENT_TIME")
MOD = sql_keyword("MOD")
MODE = sql_keyword("MODE")
NATURAL = sql_keyword("NATURAL")
NOT = sql_keyword("NOT")
NULL = sql_keyword("NULL")
NULLS = sql_keyword("NULLS")
OFFSET = sql_keyword("OFFSET")
OJ = sql_keyword("OJ")
ON = sql_keyword("ON")
ONLY = sql_keyword("ONLY")
OR = sql_keyword("OR")
ORDER = sql_keyword("ORDER")
OUTER = sql_keyword("OUTER")
PARTITION = sql_keyword("PARTITION")
PROCEDURE = sql_keyword("PROCEDURE")
QUERY = sql_keyword("QUERY")
REGEXP = sql_keyword("REGEXP")
RESTRICT = sql_keyword("RESTRICT")
RIGHT = sql_keyword("RIGHT")
ROLLUP = sql_keyword("ROLLUP")
ROW = sql_keyword("ROW")
SELECT = sql_keyword("SELECT")
SET = sql_keyword("SET")
SOUNDS = sql_keyword("SOUNDS")
SQL_BIG_RESULT = sql_keyword("SQL_BIG_RESULT")
SQL_BUFFER_RESULT = sql_keyword("SQL_BUFFER_RESULT")
SQL_CACHE = sql_keyword("SQL_CACHE")
SQL_CALC_FOUND_ROWS = sql_keyword("SQL_CALC_FOUND_ROWS")
SQL_NO_CACHE = sql_keyword("SQL_NO_CACHE")
SQL_SMALL_RESULT = sql_keyword("SQL_SMALL_RESULT")
STRAIGHT_JOIN = sql_keyword("STRAIGHT_JOIN")
TABLE = sql_keyword("TABLE")
TABLESPACE = sql_keyword("TABLESPACE")
THEN = sql_keyword("THEN")
TIME = sql_keyword("TIME")
TIMESTAMP = sql_keyword("TIMESTAMP")
TRUE = sql_keyword("TRUE")
UNION = sql_keyword("UNION")
UPDATE = sql_keyword("UPDATE")
USE = sql_keyword("USE")
USING = sql_keyword("USING")
UNKNOWN = sql_keyword("UNKNOWN")
VALUES = sql_keyword("VALUES")
WITH = sql_keyword("WITH")
WHEN = sql_keyword("WHEN")
WHERE = sql_keyword("WHERE")
XOR = sql_keyword("XOR")


all_keywords = """
    AGAINST ALL ALTER AND ASC AS
    BETWEEN BINARY BOOLEAN BY
    CASE CASCADE COLLATE CREATE CROSS
    DATETIME DATE DELETE DESC DISTINCTROW DISTINCT DIV DROP
    ELSE END ESCAPE EXISTS EXPANSION
    FALSE FIRST FORCE FOR FROM
    GROUP
    HAVING HIGH_PRIORITY
    IGNORE INDEX INNER INSERT INTERVAL INTO IN IS
    JOIN
    KEY
    LANGUAGE LAST LEFT LIKE LIMIT
    MATCH MAX_STATEMENT_TIME MODE MOD
    NATURAL NOT NULLS NULL
    OFFSET OJ ONLY ON ORDER OR OUTER
    PARTITION PROCEDURE
    QUERY
    REGEXP RESTRICT RIGHT ROLLUP ROW
    SELECT SET SOUNDS SQL_BIG_RESULT SQL_BUFFER_RESULT
        SQL_CACHE SQL_CALC_FOUND_ROWS SQL_NO_CACHE SQL_SMALL_RESULT
        STRAIGHT_JOIN
    TABLESPACE TABLE THEN TIMESTAMP TIME TRUE
    UNION UPDATE USE USING UNKNOWN
    VALUES
    WITH WHEN WHERE
    XOR
"""
keyword = make_words_regex(all_keywords, caseless=True, name="keyword")

COMMA = Literal(",")
LPAR = Literal("(")
RPAR = Literal(")")


# -----------------------------------------------------------------------------
# Literals
# -----------------------------------------------------------------------------
# http://dev.mysql.com/doc/refman/5.7/en/literals.html
# http://dev.mysql.com/doc/refman/5.7/en/date-and-time-literals.html
boolean_literal = make_words_regex("TRUE FALSE", name="boolean")
binary_literal = Regex("(b'[01]+')|(0b[01]+)").setName("binary")
hexadecimal_literal = Regex("(X'[0-9a-fA-F]+')|"
                            "(x'[0-9a-fA-F]+')|"
                            "(0x[0-9a-fA-F]+)").setName("hex")
numeric_literal = Regex(r"[+-]?\d+(\.\d*)?([eE][+-]?\d+)?")
integer = Regex(r"[+-]?\d+")  # ... simple integer
sql_keyword_literal = make_words_regex("NULL TRUE FALSE UNKNOWN",
                                       name="boolean")

string_value_singlequote = QuotedString(
    quoteChar="'", escChar="\\", escQuote="''", unquoteResults=False
).setName("single_q_string")
string_value_doublequote = QuotedString(
    quoteChar='"', escChar='\\', escQuote='""', unquoteResults=False
).setName("double_q_string")
string_literal = (
    string_value_singlequote | string_value_doublequote
).setName("string")

# http://dev.mysql.com/doc/refman/5.7/en/date-and-time-literals.html
RE_YEAR = r"([\d]{4}|[\d]{2})"
RE_MONTH = bracket(r"[0][1-9]|[1][0-2]")  # 01 to 12
RE_DAY = bracket(r"[0][0-1]|[1-2][1-9]|[3][0-1]")  # 01 to 31
RE_HOUR = bracket(r"[0-1][0-9]|[2][0-3]")  # 00 to 23
RE_MINUTE = bracket(r"[0-5][0-9]")  # 00 to 59
RE_SECOND = RE_MINUTE
FRACTIONAL_SECONDS = bracket(r"(.[0-9]{1,6})?")  # dot then from 1-6 digits
PUNCTUATION = r"[_,!@#$%&*;\\|:./\^-]"
DATESEP = bracket(r"{}?".format(PUNCTUATION))  # "any punctuation"... or none
DTSEP = bracket(r"[T ]?")  # T or none
TIMESEP = bracket(r"[:]?")  # can also be none

DATE_CORE_REGEX_STR = (
    "({year}{sep}{month}{sep}{day})".format(
        year=RE_YEAR,
        month=RE_MONTH,
        day=RE_DAY,
        sep=DATESEP
    )
)
DATE_LITERAL_REGEX_STR = single_quote(DATE_CORE_REGEX_STR)
TIME_CORE_REGEX_STR = (
    bracket(
        "({hour}{sep}{minute}{sep}{second}{fraction})|"
        "({hour}{sep}{minute})".format(
            hour=RE_HOUR,
            minute=RE_MINUTE,
            second=RE_SECOND,
            fraction=FRACTIONAL_SECONDS,
            sep=TIMESEP
        )
    )
)
TIME_LITERAL_REGEX_STR = single_quote(TIME_CORE_REGEX_STR)
DATETIME_CORE_REGEX_STR = "({date}{sep}{time})".format(
    date=DATE_CORE_REGEX_STR,
    sep=DTSEP,
    time=TIME_CORE_REGEX_STR)
DATETIME_LITERAL_REGEX_STR = single_quote(DATETIME_CORE_REGEX_STR)

date_string = Regex(DATE_LITERAL_REGEX_STR)
time_string = Regex(TIME_LITERAL_REGEX_STR)
datetime_string = Regex(DATETIME_LITERAL_REGEX_STR)
datetime_literal = (
    (Optional(DATE | 'd') + date_string) |
    (Optional(TIME | 't') + time_string) |
    (Optional(TIMESTAMP | 'ts') + datetime_string)
).setName("datetime")

literal_value = (
    numeric_literal |
    boolean_literal |
    binary_literal |
    hexadecimal_literal |
    sql_keyword_literal |
    string_literal |
    datetime_literal
).setName("literal_value")

# https://dev.mysql.com/doc/refman/5.7/en/date-and-time-functions.html#function_date-add  # noqa
time_unit = make_words_regex(
    "MICROSECOND SECOND MINUTE HOUR DAY WEEK MONTH QUARTER YEAR"
    " SECOND_MICROSECOND"
    " MINUTE_MICROSECOND MINUTE_SECOND"
    " HOUR_MICROSECOND HOUR_SECOND HOUR_MINUTE"
    " DAY_MICROSECOND DAY_SECOND DAY_MINUTE DAY_HOUR"
    " YEAR_MONTH",
    caseless=True,
    name="time_unit"
)

# -----------------------------------------------------------------------------
# Forward declarations
# -----------------------------------------------------------------------------

expr = Forward()
select_statement = Forward()

# -----------------------------------------------------------------------------
# identifier
# -----------------------------------------------------------------------------
# http://dev.mysql.com/doc/refman/5.7/en/identifiers.html
bare_identifier_word = make_regex_except_words(
    r"\b[a-zA-Z0-9$_]*\b",
    all_keywords,
    caseless=True,
    name="bare_identifier_word"
)
identifier = (
    bare_identifier_word |
    QuotedString(quoteChar="`", unquoteResults=False)
).setName("identifier")
# http://dev.mysql.com/doc/refman/5.7/en/charset-collate.html
collation_name = identifier.copy()
column_name = identifier.copy()
column_alias = identifier.copy()
table_name = identifier.copy()
table_alias = identifier.copy()
index_name = identifier.copy()
function_name = identifier.copy()
parameter_name = identifier.copy()
database_name = identifier.copy()
partition_name = identifier.copy()

no_dot = NotAny('.')
table_spec = (
    Combine(database_name + '.' + table_name + no_dot) |
    table_name + no_dot
).setName("table_spec")
column_spec = (
    Combine(database_name + '.' + table_name + '.' + column_name + no_dot) |
    Combine(table_name + '.' + column_name + no_dot) |
    column_name + no_dot
).setName("column_spec")

# http://dev.mysql.com/doc/refman/5.7/en/expressions.html
bind_parameter = Literal('?')

# http://dev.mysql.com/doc/refman/5.7/en/user-variables.html
variable = Regex(r"@[a-zA-Z0-9\.$_]+").setName("variable")

# http://dev.mysql.com/doc/refman/5.7/en/functions.html
argument_list = (
    delimitedList(expr).setName("arglist").setParseAction(', '.join)
)
# ... we don't care about sub-parsing the argument list, so use combine=True
# or setParseAction: http://stackoverflow.com/questions/37926516
function_call = Combine(function_name + LPAR) + argument_list + RPAR

# http://dev.mysql.com/doc/refman/5.7/en/partitioning-selection.html
partition_list = (
    LPAR + delim_list(partition_name, combine=True) + RPAR
).setName("partition_list")

# http://dev.mysql.com/doc/refman/5.7/en/index-hints.html
index_list = delim_list(index_name, combine=False)
# ... see pyparsing_bugtest_delimited_list_combine
index_hint = (
    (
        USE + (INDEX | KEY) +
        Optional(FOR + (JOIN | (ORDER + BY) | (GROUP + BY))) +
        LPAR + Optional(index_list) + RPAR
    ) |
    (
        IGNORE + (INDEX | KEY) +
        Optional(FOR + (JOIN | (ORDER + BY) | (GROUP + BY))) +
        LPAR + index_list + RPAR
    ) |
    (
        FORCE + (INDEX | KEY) +
        Optional(FOR + (JOIN | (ORDER + BY) | (GROUP + BY))) +
        LPAR + index_list + RPAR
    )
)
index_hint_list = delim_list(index_hint, combine=True).setName(
    "index_hint_list")

# -----------------------------------------------------------------------------
# CASE
# -----------------------------------------------------------------------------
# NOT THIS: https://dev.mysql.com/doc/refman/5.7/en/case.html
# THIS: https://dev.mysql.com/doc/refman/5.7/en/control-flow-functions.html#operator_case  # noqa
case_expr = (
    (
        CASE + expr +
        OneOrMore(WHEN + expr + THEN + expr) +
        Optional(ELSE + expr) +
        END
    ) | (
        CASE +
        OneOrMore(WHEN + expr + THEN + expr) +
        Optional(ELSE + expr) +
        END
    )
).setName("case_expr")

# -----------------------------------------------------------------------------
# MATCH
# -----------------------------------------------------------------------------
# https://dev.mysql.com/doc/refman/5.7/en/fulltext-search.html#function_match
search_modifier = (
    (IN + NATURAL + LANGUAGE + MODE + Optional(WITH + QUERY + EXPANSION)) |
    (IN + BOOLEAN + MODE) |
    (WITH + QUERY + EXPANSION)
)
match_expr = (
    MATCH + LPAR + delim_list(column_spec) + RPAR +
    AGAINST + LPAR + expr + Optional(search_modifier) + RPAR
).setName("match_expr")

# -----------------------------------------------------------------------------
# Expressions
# -----------------------------------------------------------------------------
# http://dev.mysql.com/doc/refman/5.7/en/expressions.html
# https://pyparsing.wikispaces.com/file/view/select_parser.py
# http://dev.mysql.com/doc/refman/5.7/en/operator-precedence.html
expr_term = (
    INTERVAL + expr + time_unit |
    # "{" + identifier + expr + "}" |  # see MySQL notes; antique ODBC syntax
    Optional(EXISTS) + LPAR + select_statement + RPAR |
    # ... e.g. mycol = EXISTS(SELECT ...)
    # ... e.g. mycol IN (SELECT ...)
    LPAR + delim_list(expr) + RPAR |
    # ... e.g. mycol IN (1, 2, 3)
    case_expr |
    match_expr |
    bind_parameter |
    variable |
    function_call |
    literal_value |
    column_spec  # not just identifier
)
UNARY_OP, BINARY_OP, TERNARY_OP = 1, 2, 3
expr << infixNotation(expr_term, [
    # Having lots of operations in the list here SLOWS IT DOWN A LOT.
    # Just combine them into an ordered list.
    (BINARY | COLLATE | oneOf('! - + ~'), UNARY_OP, opAssoc.RIGHT),
    (
        oneOf('^ * / %') | DIV | MOD |
        oneOf('+ - << >> & | = <=> >= > <= < <> !=') |
        (IS + Optional(NOT)) | LIKE | REGEXP | (Optional(NOT) + IN) |
        (SOUNDS + LIKE),  # RNC; presumably at same level as LIKE
        BINARY_OP,
        opAssoc.LEFT
    ),
    ((BETWEEN, AND), TERNARY_OP, opAssoc.LEFT),
    # CASE handled above (hoping precedence is not too much of a problem)
    (NOT, UNARY_OP, opAssoc.RIGHT),
    (AND | '&&' | XOR | OR | '||' | ':=', BINARY_OP, opAssoc.LEFT),
], lpar=LPAR, rpar=RPAR)
# ignores LIKE [ESCAPE]

# -----------------------------------------------------------------------------
# SELECT
# -----------------------------------------------------------------------------
# http://dev.mysql.com/doc/refman/5.7/en/select.html
"""
SELECT
    [ALL | DISTINCT | DISTINCTROW ]
      [HIGH_PRIORITY]
      [MAX_STATEMENT_TIME = N]
      [STRAIGHT_JOIN]
      [SQL_SMALL_RESULT] [SQL_BIG_RESULT] [SQL_BUFFER_RESULT]
      [SQL_CACHE | SQL_NO_CACHE] [SQL_CALC_FOUND_ROWS]
    select_expr [, select_expr ...]
    [FROM table_references
      [PARTITION partition_list]
    [WHERE where_condition]
    [GROUP BY {col_name | expr | position}
      [ASC | DESC], ... [WITH ROLLUP]]
    [HAVING where_condition]
    [ORDER BY {col_name | expr | position}
      [ASC | DESC], ...]
    [LIMIT {[offset,] row_count | row_count OFFSET offset}]

... ignore below here...

    [PROCEDURE procedure_name(argument_list)]
    [INTO OUTFILE 'file_name'
        [CHARACTER SET charset_name]
        export_options
      | INTO DUMPFILE 'file_name'
      | INTO var_name [, var_name]]
    [FOR UPDATE | LOCK IN SHARE MODE]]
"""

compound_operator = UNION + Optional(ALL | DISTINCT)
# no INTERSECT or EXCEPT in MySQL?

ordering_term = (
    expr + Optional(COLLATE + collation_name) + Optional(ASC | DESC)
)
# ... COLLATE can appear in lots of places;
# http://dev.mysql.com/doc/refman/5.7/en/charset-collate.html

join_constraint = Optional(Group(  # join_condition in MySQL grammar
    (ON + expr) |
    (USING + LPAR + delim_list(column_name) + RPAR)
))

# http://dev.mysql.com/doc/refman/5.7/en/join.html
join_op = Group(
    COMMA |
    STRAIGHT_JOIN |
    NATURAL + (Optional(LEFT | RIGHT) + Optional(OUTER)) + JOIN |
    (INNER | CROSS) + JOIN |
    Optional(LEFT | RIGHT) + Optional(OUTER) + JOIN
    # ignores antique ODBC "{ OJ ... }" syntax
)

join_source = Forward()
single_source = (
    (
        table_spec.copy().setResultsName("from_tables", listAllMatches=True) +
        Optional(PARTITION + partition_list) +
        Optional(Optional(AS) + table_alias) +
        Optional(index_hint_list)
    ) |
    (select_statement + Optional(AS) + table_alias) +
    (LPAR + join_source + RPAR)
)
join_source << Group(
    single_source + ZeroOrMore(join_op + single_source + join_constraint)
)("join_source")
# ... must have a Group to append to it later, it seems
# ... but name it "join_source" here, or it gets enclosed in a further list
#     when you name it later

result_column = (
    '*' |
    Combine(table_name + '.' + '*') |
    column_spec |
    expr + Optional(Optional(AS) + column_alias)
).setResultsName("select_columns", listAllMatches=True)

select_core = (
    SELECT +
    Group(Optional(ALL | DISTINCT | DISTINCTROW))("select_specifier") +
    Optional(HIGH_PRIORITY) +
    Optional(MAX_STATEMENT_TIME + '=' + integer) +
    Optional(STRAIGHT_JOIN) +
    Optional(SQL_SMALL_RESULT) +
    Optional(SQL_BIG_RESULT) +
    Optional(SQL_BUFFER_RESULT) +
    Optional(SQL_CACHE | SQL_NO_CACHE) +
    Optional(SQL_CALC_FOUND_ROWS) +
    Group(delim_list(result_column))("select_expression") +
    Optional(
        FROM + join_source +
        Optional(PARTITION + partition_list) +
        Group(Optional(WHERE + Group(expr)("where_expr")))("where_clause") +
        Optional(
            GROUP + BY +
            delim_list(ordering_term +
                       Optional(ASC | DESC))("group_by_term") +
            Optional(WITH + ROLLUP)
        ) +
        Optional(HAVING + expr("having_expr"))
    )
)

select_statement << (
    select_core +
    ZeroOrMore(compound_operator + select_core) +
    Optional(
        ORDER + BY +
        delim_list(ordering_term +
                   Optional(ASC | DESC))("order_by_terms")
    ) +
    Optional(LIMIT + (
        (Optional(integer("offset") + COMMA) + integer("row_count")) |
        (integer("row_count") + OFFSET + integer("offset")))
    )
    # PROCEDURE ignored
    # rest ignored
)

# -----------------------------------------------------------------------------
# Comments
# -----------------------------------------------------------------------------
# http://dev.mysql.com/doc/refman/5.7/en/comments.html
ansi_comment = Literal("--") + restOfLine
bash_comment = Literal("#") + restOfLine
comment = (ansi_comment | bash_comment | cStyleComment)

# -----------------------------------------------------------------------------
# Grammar (here, just SELECT)
# -----------------------------------------------------------------------------
select_statement.ignore(comment)
sql_grammar = select_statement


# =============================================================================
# sqlparse tests
# =============================================================================

# def parsed_from_text(sql):
#     statements = sqlparse.parse(sql)
#     if not statements:
#         return None
#     first_statement = statements[0]
#     return first_statement


# def formatted_from_parsed(statement):
#     sql = str(statement)
#     return sqlparse.format(sql, reindent=True)


# =============================================================================
# pyparsing test
# =============================================================================

# See also: parser.runTests()

def standardize_whitespace(text: str) -> str:
    return " ".join(text.split())


def test_succeed(parser: ParserElement,
                 text: str,
                 target: str = None,
                 skip_target: bool = True,
                 show_raw: bool = False,
                 verbose: bool = True) -> None:
    if target is None:
        target = text
    try:
        p = parser.parseString(text)
        log.debug("Success: {} -> {}".format(text, text_from_parsed(p)))
        if show_raw:
            log.debug("... raw: {}".format(p))
        if verbose:
            log.debug("... dump:\n{}".format(p.dump()))
    except ParseException as exception:
        log.debug("Failure on: {} [parser: {}]".format(text, parser))
        print(statement_and_failure_marker(text, exception))
        raise
    if not skip_target:
        intended = standardize_whitespace(target)
        actual = standardize_whitespace(text_from_parsed(p))
        if intended != actual:
            raise ValueError(
                "Failure on: {} -> {} (should have been: {})"
                " [parser: {}] [as list: {}]".format(
                    text, actual, intended, parser, repr(p.asList())))


def test_fail(parser: ParserElement, text: str) -> None:
    try:
        p = parser.parseString(text)
        raise ValueError(
            "Succeeded erroneously: {} -> {} [raw: {}] [parser: {}]".format(
                text, text_from_parsed(p), p, parser))
    except ParseException:
        log.debug("Correctly failed: {}".format(text))


def test_select(text: str) -> None:
    test_succeed(select_statement, text, verbose=True)


def test(parser: ParserElement, text: str) -> None:
    print("STATEMENT:\n{}".format(text))

    # scanned = parser.scanString(text)
    # for tokens, start, end in scanned:
    #     print(start, end, tokens)

    try:
        tokens = parser.parseString(text)
        print(tokens.asXML())
        # print(tokens.dump())
        # print(text_from_parsed(tokens))
        print("tokens = {}".format(tokens))
        d = tokens.asDict()
        for k, v in d.items():
            print("tokens.{} = {}".format(k, v))
        for c in tokens.columns:
            print("column: {}".format(c))
    except ParseException as err:
        print(" "*err.loc + "^\n" + err.msg)
        print(err)
    print()


def flatten(*args):
    for x in args:
        if isinstance(x, (list, tuple)):
            for y in flatten(*x):
                yield y
        else:
            yield x


def text_from_parsed(parsetree: ParseResults,
                     formatted: bool = True,
                     indent_width: int = 4) -> str:
    nested_list = parsetree.asList()
    flattened_list = flatten(nested_list)
    plain_str = " ".join(flattened_list)
    if not formatted:
        return plain_str

    # Forget my feeble efforts for now (below) and use sqlparse:
    return sqlparse.format(plain_str, reindent=True,
                           indent_width=indent_width)

    # result = ""
    # newline_before = ["SELECT", "FROM", "WHERE", "GROUP"]
    # indent_after = ["SELECT", "FROM", "WHERE", "GROUP"]
    # no_space_before = [",", ")"]
    # no_space_after = ["("]
    # at_line_start = True
    # indent = 0
    #
    # def start_new_line(indent_change=0):
    #     nonlocal at_line_start
    #     nonlocal indent
    #     nonlocal result
    #     if at_line_start:
    #         return
    #     indent += indent_change
    #     result += "\n" + " " * indent * indent_spaces
    #     at_line_start = True
    #
    # previous_item = None
    # for index, item in enumerate(flattened_list):
    #     if item in newline_before:
    #         start_new_line(indent_change=-1)
    #     leading_space = (
    #         not at_line_start and
    #         item not in no_space_before and
    #         previous_item not in no_space_after
    #     )
    #     if leading_space:
    #         result += " "
    #     result += item
    #     at_line_start = False
    #     if item in indent_after:
    #         start_new_line(indent_change=1)
    #     previous_item = item
    # return result


def statement_and_failure_marker(text: str,
                                 parse_exception: ParseException) -> str:
    output_lines = []
    for line_index, line in enumerate(text.split("\n")):
        output_lines.append(line)
        if parse_exception.lineno == line_index + 1:
            output_lines.append("-" * (parse_exception.col - 1) + "^")
    return "\n".join(output_lines)


# =============================================================================
# Tests
# =============================================================================

# noinspection PyUnboundLocalVariable
def pyparsing_bugtest_delimited_list_combine(fix_problem: bool = True) -> None:
    if fix_problem:
        # noinspection PyPep8Naming,PyShadowingNames
        delimitedList = delim_list
    word = Word(alphanums)
    word_list_no_combine = delimitedList(word, combine=False)
    word_list_combine = delimitedList(word, combine=True)
    print(word_list_no_combine.parseString('one, two'))  # ['one', 'two']
    print(word_list_no_combine.parseString('one,two'))  # ['one', 'two']
    print(word_list_combine.parseString('one, two'))  # ['one']: ODD ONE OUT
    print(word_list_combine.parseString('one,two'))  # ['one,two']


def unit_tests(test_expr: bool = False) -> None:
    # -------------------------------------------------------------------------
    # pyparsing tests
    # -------------------------------------------------------------------------
    log.info("Testing pyparsing elements")
    regexp_for = Regex(r"\bfor\b", flags=re.IGNORECASE)
    test_succeed(regexp_for, "for")
    test_fail(regexp_for, "blandford")
    test_fail(regexp_for, "upfor")
    test_fail(regexp_for, "forename")

    # -------------------------------------------------------------------------
    # Literals
    # -------------------------------------------------------------------------
    log.info("Testing boolean_literal")
    test_succeed(boolean_literal, "TRUE")
    test_succeed(boolean_literal, "FALSE")
    test_fail(boolean_literal, "blah")

    log.info("Testing binary_literal")
    test_succeed(binary_literal, "b'010101'")
    test_succeed(binary_literal, "0b010101")

    log.info("Testing hexadecimal_literal")
    test_succeed(hexadecimal_literal, "X'12fac'")
    test_succeed(hexadecimal_literal, "x'12fac'")
    test_succeed(hexadecimal_literal, "0x12fac")

    log.info("Testing numeric_literal")
    test_succeed(numeric_literal, "45")
    test_succeed(numeric_literal, "+45")
    test_succeed(numeric_literal, "-45")
    test_succeed(numeric_literal, "-45E-3")
    test_succeed(numeric_literal, "-45E3")
    test_succeed(numeric_literal, "-45.32")
    test_succeed(numeric_literal, "-45.32E6")

    log.info("Testing string_value_singlequote")
    test_succeed(string_value_singlequote, "'single-quoted string'")
    log.info("Testing string_value_doublequote")
    test_succeed(string_value_doublequote, '"double-quoted string"')
    log.info("Testing string_literal")
    test_succeed(string_literal, "'single-quoted string'")
    test_succeed(string_literal, '"double-quoted string"')

    log.info("Testing date_string")
    test_succeed(date_string, single_quote("2015-04-14"))
    test_succeed(date_string, single_quote("20150414"))
    log.info("Testing time_string")
    test_succeed(time_string, single_quote("15:23"))
    test_succeed(time_string, single_quote("1523"))
    test_succeed(time_string, single_quote("15:23:00"))
    test_succeed(time_string, single_quote("15:23:00.1"))
    test_succeed(time_string, single_quote("15:23:00.123456"))
    log.info("Testing datetime_string")
    test_succeed(datetime_string, single_quote("2015-04-14 15:23:00.123456"))
    test_succeed(datetime_string, single_quote("2015-04-14T15:23:00.123456"))

    log.info("Testing literal")
    test_succeed(literal_value, "NULL")

    log.info("Testing time_unit")
    test_succeed(time_unit, "MICROSECOND")
    test_succeed(time_unit, "year_month")

    # -------------------------------------------------------------------------
    # Identifiers
    # -------------------------------------------------------------------------

    log.info("Testing FOR")
    print(FOR.pattern)
    test_succeed(FOR, "for")
    test_fail(FOR, "thingfor")  # shouldn't match FOR
    test_fail(FOR, "forename")  # shouldn't match FOR

    log.info("Testing keyword")
    print(keyword.pattern)
    test_succeed(keyword, "TABLE")
    test_fail(keyword, "thingfor")  # shouldn't match FOR
    test_fail(keyword, "forename")  # shouldn't match FOR

    log.info("Testing bare_identifier_word")
    test_succeed(bare_identifier_word, "blah")
    test_fail(bare_identifier_word, "FROM")
    test_succeed(bare_identifier_word, "forename")

    log.info("Testing identifier")
    test_succeed(identifier, "blah")
    test_succeed(identifier, "idx1")
    test_succeed(identifier, "idx2")
    test_succeed(identifier, "a")
    test_succeed(identifier, "`FROM`")
    test_succeed(identifier, "`SELECT FROM`")
    log.info("... done")

    log.info("Testing table_spec")
    test_succeed(table_spec, "mytable")
    test_succeed(table_spec, "mydb.mytable")
    test_succeed(table_spec, "mydb.`my silly table`")
    test_fail(table_spec, "mydb . mytable")
    test_fail(table_spec, "mydb.mytable.mycol")

    log.info("Testing column_spec")
    test_succeed(column_spec, "mycol")
    test_succeed(column_spec, "forename")
    test_succeed(column_spec, "mytable.mycol")
    test_succeed(column_spec, "t1.a")
    test_succeed(column_spec, "`my silly table`.`my silly column`")
    test_succeed(column_spec, "mydb.mytable.mycol")
    test_fail(column_spec, "mydb . mytable . mycol")

    log.info("Testing variable")
    test_succeed(variable, "@myvar")

    log.info("Testing argument_list")
    test_succeed(argument_list, "@myvar, 5")

    log.info("Testing function_call")
    test_succeed(function_call, "myfunc(@myvar, 5)")

    log.info("Testing index_list")
    test_succeed(index_list, "idx1, idx2")

    log.info("Testing index_hint")
    test_succeed(index_hint, "USE INDEX FOR JOIN (idx1, idx2)")

    # -------------------------------------------------------------------------
    # Expressions
    # -------------------------------------------------------------------------

    log.info("Testing case_expr")
    test_succeed(case_expr, """
        CASE v
          WHEN 2 THEN x
          WHEN 3 THEN y
          ELSE -99
        END
    """)

    log.info("Testing match_expr")
    test_succeed(match_expr, """
         MATCH (content_field)
         AGAINST('+keyword1 +keyword2' IN BOOLEAN MODE)
    """)

    log.info("Testing expr_term")
    test_succeed(expr_term, "5")
    test_succeed(expr_term, "5.12")
    test_succeed(expr_term, "'string'")
    test_succeed(expr_term, "mycol")
    test_succeed(expr_term, "myfunc(myvar, 8)")
    test_succeed(expr_term, "INTERVAL 5 MICROSECOND")
    test_succeed(expr_term, "(SELECT 1)")
    test_succeed(expr_term, "(1, 2, 3)")

    if test_expr:
        log.info("Testing expr")
        test_succeed(expr, "5")
        test_succeed(expr, "a")
        test_succeed(expr, "mycol1 || mycol2")
        test_succeed(expr, "+mycol")
        test_succeed(expr, "-mycol")
        test_succeed(expr, "~mycol")
        test_succeed(expr, "!mycol")
        test_succeed(expr, "a | b")
        test_succeed(expr, "a & b")
        test_succeed(expr, "a << b")
        test_succeed(expr, "a >> b")
        test_succeed(expr, "a + b")
        test_succeed(expr, "a - b")
        test_succeed(expr, "a * b")
        test_succeed(expr, "a / b")
        test_succeed(expr, "a DIV b")
        test_succeed(expr, "a MOD b")
        test_succeed(expr, "a % b")
        test_succeed(expr, "a ^ b")
        test_succeed(expr, "a ^ (b + (c - d) / e)")
        test_succeed(expr, "a NOT IN (SELECT 1)")
        test_succeed(expr, "a IN (1, 2, 3)")
        test_succeed(expr, "a IS NULL")
        test_succeed(expr, "a IS NOT NULL")
        test_fail(expr, "IS NULL")
        test_fail(expr, "IS NOT NULL")
        test_succeed(expr, "(a * (b - 3)) > (d - 2)")

    log.info("Testing join_op")
    test_succeed(join_op, ",")
    test_succeed(join_op, "INNER JOIN")

    log.info("Testing join_source")
    test_succeed(join_source, "a")
    test_succeed(join_source, "a INNER JOIN b")
    test_succeed(join_source, "a, b")

    log.info("Testing result_column")
    test_succeed(result_column, "t1.a")

    log.info("Testing select_statement")
    test_select("SELECT a, b FROM c")  # no WHERE
    test_select("SELECT a, b FROM c WHERE d > 5 AND e = 4")
    test_select("SELECT a, b FROM c INNER JOIN f WHERE d > 5 AND e = 4")
    test_select("SELECT t1.a, t1.b FROM t1 WHERE t1.col1 > 5")
    test_select("SELECT t1.a, t1.b FROM t1 WHERE t1.col1 IN (1, 2, 3)")


# =============================================================================
# main
# =============================================================================

def main() -> None:
    # blank = ''
    # p_blank = parsed_from_text(blank)
    # p_sql1 = parsed_from_text(sql1)
    # print(formatted_from_parsed(p_blank))
    # print(formatted_from_parsed(p_sql1))

    # pyparsing_bugtest_delimited_list_combine()
    unit_tests(test_expr=True)


if __name__ == '__main__':
    main_only_quicksetup_rootlogger()
    main()
